"""klayout_pyxs.py

"""
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A feasibility study for a cross section generation using
# boolean operations. See "cmos.pyxs" for a brief description of the
# commands available and some examples.
# TODO: use a much smaller dbu for the simulation to have a really small delta
# the paths used for generating the masks are somewhat too thick
# TODO: the left and right areas are not treated correctly
import math
import os
import re

from klayout_pyxs import HAS_PYA, Box, Edge, Point, Polygon
from klayout_pyxs.compat import range, zip

# from importlib import reload
# try:
#     reload(klayout_pyxs)
#     reload(klayout_pyxs.utils)
#     reload(klayout_pyxs.geometry_2d)
# except:
#     pass


if HAS_PYA:
    # Imports for KLayout plugin
    from klayout_pyxs import Application
    from klayout_pyxs import MessageBox
    from klayout_pyxs import Action
    from klayout_pyxs import FileDialog

from klayout_pyxs.geometry_2d import EP, LayoutData, MaskData, MaterialData, ep
from klayout_pyxs.layer_parameters import string_to_layer_info
from klayout_pyxs.utils import info, int_floor, make_iterable, print_info

info("Module klayout_pyxs.pyxs_lib.py reloaded")


class XSectionGenerator:
    """The main class that creates a cross-section file"""

    def __init__(self, file_name):
        """
        Parameters
        ----------
        file_name : str
        """
        # TODO: adjust this path:
        self._file_name = file_name
        self._file_path = os.path.split(file_name)[0]
        self._lyp_file = None
        self._ep = ep
        self._flipped = False
        self._air = None  # type: MaterialData
        self._air_below = None  # type: MaterialData
        self._delta = None
        self._extend = None
        self._below = None
        self._depth = None
        self._height = None

        self._is_target_layout_created = False
        self._hide_png_save_error = False

        self._output_all_parameters = {
            "save_png": False,
            "output_layers": None,
            "png_path": None,
        }

    def layer(self, layer_spec):
        """Fetches an input layer from the original layout.

        Parameters
        ----------
        layer_spec : str

        Returns
        -------
        ld : LayerData

        """
        ld = LayoutData([], self)  # empty
        # collect shapes from the corresponding layer into ld._polygons
        ld.load(
            self._layout,
            self._cell,
            self._line_dbu.bbox().enlarge(Point(self._extend, self._extend)),
            layer_spec,
        )
        return ld

    @print_info(False)
    def mask(self, layer_data):
        """Designates the layout_data object as a litho pattern (mask).

        This is the starting point for structured grow or etch operations.

        Parameters
        ----------
        layer_data : LayoutData

        Returns
        -------
        res : klayout_pyxs_repo.klayout_pyxs.geometry_2d.MaterialData
        """
        crossing_points = []

        info(f"    layer_data: {layer_data}")

        info(f"    n polygons in layer_data: {layer_data.n_poly}")

        for polygon in layer_data.data:
            info(f"    polygon: {polygon}")
            for edge_dbu in polygon.each_edge():
                info(f"        edge: {edge_dbu}")
                if self._line_dbu.crossed_by(edge_dbu):
                    info("        crosses!")

                if self._line_dbu.crossed_by(edge_dbu) and (
                    self._line_dbu.side_of(edge_dbu.p1) > 0
                    or self._line_dbu.side_of(edge_dbu.p2) > 0
                ):
                    info("        inside if")
                    # compute the crossing point of "edge" and "line" in
                    # database units
                    # confine the point to the length of the line
                    z = (
                        float(edge_dbu.dx()) * (edge_dbu.p1.y - self._line_dbu.p1.y)
                        - float(edge_dbu.dy()) * (edge_dbu.p1.x - self._line_dbu.p1.x)
                    ) / (
                        float(edge_dbu.dx())
                        * (self._line_dbu.p2.y - self._line_dbu.p1.y)
                        - float(edge_dbu.dy())
                        * (self._line_dbu.p2.x - self._line_dbu.p1.x)
                    )
                    z = math.floor(z * self._line_dbu.length() + 0.5)
                    if z < -self._extend:
                        z = -self._extend
                    elif z > self._line_dbu.length() + self._extend:
                        z = self._line_dbu.length() + self._extend

                    v = (
                        edge_dbu.dy() * self._line_dbu.dx()
                        - edge_dbu.dx() * self._line_dbu.dy()
                    )
                    if v < 0:
                        s = -1
                    elif v == 0:
                        s = 0
                    else:
                        s = 1

                    # store that along with the orientation of the edge
                    # (+1: "enter geometry", -1: "leave geometry")
                    info(f"        appending x-point [{z}, {s}]")
                    crossing_points.append([z, s])

        # compress the crossing points by collecting all of those which
        # cut the measure line at the same position
        compressed_crossing_points = []
        last_z = None
        sum_s = 0
        crossing_points.sort()
        for z, s in crossing_points:
            if z == last_z:
                sum_s += s
            else:
                if sum_s != 0:
                    compressed_crossing_points.append([last_z, sum_s])
                last_z, sum_s = z, s

        if last_z and sum_s != 0:
            compressed_crossing_points.append([last_z, sum_s])

        # create the final intervals by selecting those crossing points which
        # denote an entry or leave point into or out of drawn geometry. This
        # basically does a merge of all drawn shapes.
        return self._xpoints_to_mask(compressed_crossing_points)

    # @property
    def air(self):
        """

        Returns
        -------
        MaterialData
        """
        return self._air

    # @property
    def bulk(self):
        """Return a material describing the wafer body

        Return
        ------
        bulk : klayout_pyxs_repo.klayout_pyxs.geometry_2d.MaterialData
        """

        return MaterialData(self._bulk.data, self)

    def output(self, layer_spec=None, layer_data=None, *args):
        """Outputs a material object to the output layout

        Can be used for a single material (layer_spec and layer_data pair),
        or for a list of materials passed through an output_layers
        dictionary.

        Parameters
        ----------
        layer_spec : str
            layer specification
        layer_data : LayoutData
        """
        # process layer_spec / layer_data pair
        if not isinstance(layer_data, LayoutData):
            raise TypeError(
                f"'output()': layer_data parameter must be a geometry object. {type(layer_data)} is given"
            )

        if not self._is_target_layout_created:
            self._create_new_layout()

        ls = string_to_layer_info(layer_spec)
        li = self._target_layout.insert_layer(ls)
        shapes = self._target_layout.cell(self._target_cell).shapes(li)

        # confine the shapes to the region of interest
        for polygon in self._ep.boolean_to_polygon(
            [Polygon(self._roi)], layer_data.data, EP.ModeAnd, True, True
        ):
            shapes.insert(polygon)

    def output_all(
        self,
        output_layers=None,
        script_globals=None,
        new_target_layout=True,
        step_name=None,
        save_png=None,
        *args,
    ):
        """Output a list of material objects to the output layout

        A list of materials is passed through an output_layers dictionary.

        Parameters
        ----------
        output_layers :  Dict[LayoutData, str] or Dict[str, str]
            Key is either LayoutData instance, or a name of such instance
            instance in script_globals.
            Value is the layer specification (see layer_spec in output()).
        script_globals : Dict[str, Any]
            globals() dictionary
        new_target_layout : bool
            if True, a new target layout will be created
        step_name : str
            name extension for the newly created layout
        save_png : bool
            if True, the resulting view will be saved as an image in the
            gds file folder
        """
        if script_globals is None:
            script_globals = {}

        if new_target_layout:
            if self._is_target_layout_created:
                self._finalize_view()
            self._create_new_layout(cell_name_extension=step_name)

        if output_layers:
            ol = output_layers
        elif self._output_all_parameters["output_layers"]:
            ol = self._output_all_parameters["output_layers"]
        else:
            return None

        for ld, ls in ol.items():
            if isinstance(ld, str):
                if ld == "air":
                    a = self.air()
                    self.output(layer_spec=ls, layer_data=a)
                # elif ld in list(globals.keys()):
                #     print('yes', ld, 'in globals')
                #     self.output(layer_spec=ls,
                #                 layer_data=globals[ld])
                elif ld in list(script_globals.keys()):
                    self.output(layer_spec=ls, layer_data=script_globals[ld])
                else:
                    # skip a non-existing (yet) material
                    continue
            else:
                self.output(layer_spec=ls, layer_data=ld)

        sp = self._output_all_parameters["save_png"] if save_png is None else save_png
        if sp:
            self._finalize_view()
            if step_name:
                file_name = f"{self._cell_file_name} ({step_name}).png"
            else:
                file_name = f"{self._cell_file_name}.png"

            if self._output_all_parameters["png_path"]:
                file_name = os.path.join(
                    self._output_all_parameters["png_path"], file_name
                )
            else:
                file_name = os.path.join(self._file_path, file_name)

            try:
                self._target_view.save_image(
                    file_name,
                    self._area.width() / 100,
                    self._area.height() / 100,
                )
            except Exception as e:
                if not self._hide_png_save_error:
                    MessageBox.critical(
                        "Error",
                        f"Error saving png file {file_name}. \n\n Error: {e}. \n\nFurther error messages will not be displayed.",
                        MessageBox.b_ok(),
                    )

                    self._hide_png_save_error = True
        return None

    def output_raw(self, layer_spec, d):
        """For debugging only"""
        ls = string_to_layer_info(layer_spec)
        li = self._target_layout.insert_layer(ls)
        shapes = self._target_layout.cell(self._target_cell).shapes(li)
        shapes.insert(d)
        return shapes

    @print_info(False)
    def all(self):
        """A pseudo-mask, covering the whole wafer

        Return
        ------
        res : MaterialData
        """
        e = self._extend
        info(f"e = {e}")

        line_dbu = self._line_dbu
        info(f"line_dbu = {line_dbu}")

        res = self._xpoints_to_mask([[-e, 1], [line_dbu.length() + e, -1]])

        info(f"    all().res = {res}")
        return res

    def flip(self):
        """Start or end backside processing"""
        self._air, self._air_below = self._air_below, self._air
        self._flipped = not self._flipped

    def diffuse(self, *args, **kwargs):
        """Same as deposit()"""
        return self.all().grow(*args, **kwargs)

    def deposit(self, *args, **kwargs):
        """Deposits material as a uniform sheet.

        Equivalent to all.grow(...)

        Return
        ------
        res : MaterialData
        """
        return self.all().grow(*args, **kwargs)

    @print_info(False)
    def grow(self, *args, **kwargs):
        """Same as deposit()"""
        all = self.all()
        info(all)
        return all.grow(*args, **kwargs)

    def etch(self, *args, **kwargs):
        """Uniform etching

        Equivalent to all.etch(...)

        """
        return self.all().etch(*args, **kwargs)

    @print_info(False)
    def planarize(self, *args, **kwargs):
        """Planarization"""
        downto = None
        less = None
        to = None
        into = []

        for k, v in kwargs.items():
            if k == "downto":
                downto = make_iterable(v)
                for i in downto:
                    if not isinstance(i, MaterialData):
                        raise TypeError(
                            "'planarize' method: 'downto' expects "
                            "a material parameter or an array "
                            "of such"
                        )

            elif k == "into":
                into = make_iterable(v)
                for i in into:
                    if not isinstance(i, MaterialData):
                        raise TypeError(
                            "'planarize' method: 'into' expects "
                            "a material parameter or an array "
                            "of such"
                        )
            elif k == "less":
                less = int_floor(0.5 + float(v) / self.dbu)
            elif k == "to":
                to = int_floor(0.5 + float(v) / self.dbu)

        if not into:
            raise ValueError("'planarize' requires an 'into' argument")

        info(f"   downto = {downto}")
        info(f"   less = {less}")
        info(f"   to = {to}")
        info(f"   into = {into}")

        if downto:
            downto_data = None
            if len(downto) == 1:
                downto_data = downto[0].data
            else:
                for i in downto:
                    if len(downto_data) == 0:
                        downto_data = i.data
                    else:
                        downto_data = self._ep.boolean_p2p(
                            i.data, downto_data, EP.ModeOr
                        )

            # determine upper bound of material
            if downto_data:
                for p in downto_data:
                    yt = p.bbox().top
                    yb = p.bbox().bottom
                    to = to or yt
                    to = min([to, yt, yb]) if self._flipped else max([to, yt, yb])
                info(f"    to = {to}")

        elif into and not to:

            # determine upper bound of our material
            for i in into:
                for p in i.data:
                    yt = p.bbox().top
                    yb = p.bbox().bottom
                    to = to or yt
                    to = min([to, yt, yb]) if self._flipped else max([to, yt, yb])
        if to is not None:
            info("    to is true")
            less = less or 0
            if self._flipped:
                removed_box = Box(
                    -self._extend,
                    -self.depth_dbu - self.below_dbu,
                    self._line_dbu.length() + self._extend,
                    to + less,
                )
            else:
                removed_box = Box(
                    -self._extend,
                    to - less,
                    self._line_dbu.length() + self._extend,
                    self.height_dbu,
                )

            rem = LayoutData([], self)
            for i in into:
                rem.add(i.and_([Polygon(removed_box)]))
                i.sub([Polygon(removed_box)])

            self.air().add(rem)

            # self.air().close_gaps()

    def set_thickness_scale_factor(self, factor):
        """Configures layer thickness scale factor

        to have better proportions
        """
        self._thickness_scale_factor = factor

    def set_output_all_parameters(
        self, save_png=None, output_layers=None, png_path=None
    ):
        assert save_png is None or isinstance(save_png, bool)
        assert output_layers is None or isinstance(output_layers, dict)

        if save_png:
            self._output_all_parameters["save_png"] = save_png

        if output_layers:
            self._output_all_parameters["output_layers"] = output_layers

        if png_path:
            if not os.path.exists(png_path):
                os.makedirs(png_path)
            self._output_all_parameters["png_path"] = png_path

    @print_info(False)
    def set_delta(self, x):
        """Configures the accuracy parameter"""
        self._delta = int_floor(x / self._dbu + 0.5)
        info(f"XSG._delta set to {self._delta}")

    @print_info(False)
    def delta(self, x):
        self._delta = int_floor(x / self._dbu + 0.5)
        info(f"XSG._delta set to {self._delta}")

    @property
    def delta_dbu(self):
        return self._delta

    @print_info(False)
    def set_height(self, x):
        """Configures the height of the processing window"""
        self._height = int_floor(x / self._dbu + 0.5)
        info(f"XSG._height set to {self._height}")
        self._update_basic_regions()

    @print_info(False)
    def height(self, x):
        """Configures the height of the processing window"""
        self._height = int_floor(x / self._dbu + 0.5)
        info(f"XSG._height set to {self._height}")
        self._update_basic_regions()

    @property
    def height_dbu(self):
        return self._height

    @print_info(False)
    def depth(self, x):
        """Set the depth of the processing window
        or the wafer thickness for backside processing (see below)

        """
        self._depth = int_floor(x / self._dbu + 0.5)
        info(f"XSG._depth set to {self._depth}")
        self._update_basic_regions()

    @print_info(False)
    def set_depth(self, x):
        """Set the depth of the processing window
        or the wafer thickness for backside processing (see below)

        """
        self._depth = int_floor(x / self._dbu + 0.5)
        info(f"XSG._depth set to {self._depth}")
        self._update_basic_regions()

    @property
    def depth_dbu(self):
        return self._depth

    @print_info(False)
    def below(self, x):
        """Set the lower height of the processing window for backside processing

        Parameters
        ----------
        x : float
            depth below the wafer in um,

        """
        self._below = int_floor(x / self._dbu + 0.5)
        info(f"XSG._below set to {self._below}")
        self._update_basic_regions()

    @print_info(False)
    def set_below(self, x):
        """Set the lower height of the processing window for backside processing

        Parameters
        ----------
        x : float
            depth below the wafer in um,

        """
        self._below = int_floor(x / self._dbu + 0.5)
        info(f"XSG._below set to {self._below}")
        self._update_basic_regions()

    @property
    def below_dbu(self):
        return self._below

    def set_extend(self, x):
        """Configures the computation margin"""
        self._extend = int_floor(x / self._dbu + 0.5)
        self._update_basic_regions()

    @property
    def extend_dbu(self):
        return self._extend

    @property
    def width_dbu(self):
        """Cross-section width.

        Determined by the ruler width.
        """
        return self._line_dbu.length()

    def background(self):
        """
        Returns
        -------
        res : Box
            The extended box including the ruler.
        """
        x1 = self._line_dbu.p1.x
        y1 = self._line_dbu.p1.y
        x2 = self._line_dbu.p2.x
        y2 = self._line_dbu.p2.y
        if x2 < x1:
            x1, x2 = x2, x1

        if y2 < y1:
            y1, y2 = y2, y1

        x1 -= self._extend
        y1 -= self._extend
        x2 += self._extend
        y2 += self._extend
        return Box(
            Point(x1 - self._delta * 5, y1 - self._delta * 5),
            Point(x2 + self._delta * 5, y2 + self._delta * 5),
        )

    @property
    def dbu(self):
        return self._dbu

    def layers_file(self, lyp_file):
        """Set a .lyp layer properties file to be used on the cross-section layout"""
        self._lyp_file = lyp_file

    # The basic generation method
    def run(self, p1, p2, ruler_text=None):
        """

        Parameters
        ----------
        ruler_text : str
            identifier to be used to name a new cross-section cell

        Returns
        -------
        LayoutView
        """
        self._target_view = None
        self._target_cell_name = f"PYXS: {ruler_text}" if ruler_text else "XSECTION"
        self._cell_file_name = f"PYXS_{ruler_text}" if ruler_text else "XSECTION"

        self._setup(p1, p2)

        self._update_basic_regions()

        text = None
        try:
            with open(self._file_name) as file:
                text = file.read()
        except Exception as e:
            MessageBox.critical(
                "Error",
                f"Error reading file {self._file_name}. \n\nError: {e}",
                MessageBox.b_ok(),
            )
            return None

        if text is None:
            MessageBox.critical(
                "Error",
                f"Error reading file {self._file_name}.",
                MessageBox.b_ok(),
            )

            return None

        # prepare variables to be visible in the script
        locals_ = dir(self)
        locals_dict = {attr: getattr(self, attr) for attr in locals_ if attr[0] != "_"}
        try:
            exec(text, locals_dict)
        except Exception as e:
            # For development
            # print(e.__traceback__.)
            # print(dir(e))
            MessageBox.critical("Error", str(e), MessageBox.b_ok())
            # pass
            return None

        Application.instance().main_window().cm_lv_add_missing()  # @@@
        self._finalize_view()
        return self._target_view

    def _finalize_view(self):
        if self._target_view:
            if self._lyp_file:
                self._target_view.load_layer_props(self._lyp_file)

            self._target_view.zoom_fit()
            self._target_view.max_hier_levels = 1

    @print_info(False)
    def _xpoints_to_mask(self, iv):
        """Convert crossing points to a mask

        Parameters
        ----------
        iv : list of lists or list of tuple
            each list / tuple represents two coordinates.

        Return
        ------
        res : MaterialData
            Top of the surface for deposition
        """
        info(f"    iv = {iv})")
        s = 0
        last_s = 0
        p1 = 0
        p2 = 0

        mask_polygons = []
        for i in iv:
            z = i[0]  # first coordinate
            s += i[1]  # increase second coordinate

            if last_s <= 0 < s:  # s increased and became > 0
                p1 = z
            elif last_s > 0 >= s:  # s decreased and became < 0
                p2 = z
                poly = Polygon(Box(p1, -self._depth - self._below, p2, self._height))
                info(f"        Appending poly {poly}")
                mask_polygons.append(poly)
            last_s = s

        info(f"    mask_polys = {mask_polygons}")

        """
        air = self._air.data
        info('    air =        {}'.format(air))

        # Sizing is needed only in vertical direction, it seems
        # air_sized = self._ep.size_p2p(air, self._delta, self._delta)
        air_sized = self._ep.size_p2p(air, self._delta, self._delta)
        info('    air_sized =  {}'.format(air_sized))

        # extended air minus air
        air_border = self._ep.boolean_p2p(air_sized, air, EP.ModeANotB)
        info('    air_border = {}'.format(air_border))

        # overlap of air border and mask polygons
        mask_data = self._ep.boolean_p2p(
                air_border, mask_polygons,
                EP.ModeAnd)
        info('    mask_data  = {}'.format(mask_data))

        # info('____Creating MD from {}'.format([str(p) for p in mask_data]))
        return MaterialData(mask_data, self)
        """
        info("Before MaskData creation")
        res = MaskData(self._air.data, mask_polygons, self)
        info(f"res = {res}")
        return res

    @print_info(False)
    def _update_basic_regions(self):

        h = self._height  # height above the wafer
        d = self._depth  # thickness of the wafer
        b = self._below  # distance below the wafer

        w = self._line_dbu.length()  # length of the ruler
        e = self._extend  # extend to the sides

        self._area = Box(-e, -(d + b), w + e, h)
        self._air = MaterialData([Polygon(Box(-e, 0, w + e, h))], self)
        self._air_below = MaterialData([Polygon(Box(-e, -(d + b), w + e, -d))], self)

        self._bulk = MaterialData([Polygon(Box(-e, -d, w + e, 0))], self)
        self._roi = Box(0, -(d + b), w, h)

        info(f"    XSG._area:      {self._area}")
        info(f"    XSG._roi:       {self._roi}")
        info(f"    XSG._air:       {self._air}")
        info(f"    XSG._bulk:      {self._bulk}")
        info(f"    XSG._air_below: {self._air_below}")

    @print_info(False)
    def _setup(self, p1, p2):
        """
        Parameters
        ----------
        p1 : Point
            first point of the ruler
        p2 : Point
            second point of the ruler

        """
        # locate the layout
        app = Application.instance()
        view = app.main_window().current_view()  # LayoutView
        if not view:
            MessageBox.critical(
                "Error",
                "No view open for creating the cross-" "section from",
                MessageBox.b_ok(),
            )
            return False

        cv = view.cellview(view.active_cellview_index())  # CellView
        if not cv.is_valid():
            MessageBox.critical(
                "Error", "The selected layout is not valid", MessageBox.b_ok()
            )
            return False

        self._cv = cv  # CellView
        self._layout = cv.layout()  # Layout
        self._dbu = self._layout.dbu
        self._cell = cv.cell_index  # int

        # get the start and end points in database units and micron
        p1_dbu = Point.from_dpoint(p1 * (1.0 / self._dbu))
        p2_dbu = Point.from_dpoint(p2 * (1.0 / self._dbu))
        self._line_dbu = Edge(p1_dbu, p2_dbu)  # Edge describing the ruler

        # initialize height and depth
        self._extend = int_floor(2.0 / self._dbu + 0.5)  # 2 um in dbu
        self._delta = 10
        self._height = int_floor(2.0 / self._dbu + 0.5)  # 2 um in dbu
        self._depth = int_floor(2.0 / self._dbu + 0.5)  # 2 um in dbu
        self._below = int_floor(2.0 / self._dbu + 0.5)  # 2 um in dbu

        info(f"    XSG._dbu is:    {self._dbu}")
        info(f"    XSG._extend is: {self._extend}")
        info(f"    XSG._delta is:  {self._delta}")
        info(f"    XSG._height is: {self._height}")
        info(f"    XSG._depth is:  {self._depth}")
        info(f"    XSG._below is:  {self._below}")

        return True

    def _create_new_layout(self, cell_name_extension=None):
        if cell_name_extension:
            cell_name = f"{self._target_cell_name} ({cell_name_extension})"
        else:
            cell_name = self._target_cell_name

        # create a new layout for the output
        app = Application.instance()
        main_window = app.main_window()
        cv = main_window.create_layout(1)  # type: CellView
        cell = cv.layout().add_cell(cell_name)  # type: Cell
        self._target_view = main_window.current_view()  # type: LayoutView
        self._target_view.select_cell(cell, 0)
        self._target_layout = cv.layout()  # type: Layout
        self._target_layout.dbu = self._dbu
        self._target_cell = cell  # type: cell

        self._is_target_layout_created = True


# MENU AND ACTIONS
# ----------------
N_PYXS_SCRIPTS_MAX = 4

# pyxs_script_load_menuhandler = None
pyxs_scripts = None


class MenuHandler(Action):
    """Handler for the load .xs file action"""

    def __init__(self, title, action, shortcut=None, icon=None):
        """
        Parameters
        ----------
        title : str
        action : callable
        shortcut : str
        icon : str
        """
        self.title = title
        self._action = action
        if shortcut:
            self.shortcut = shortcut
        if icon:
            self.icon = icon

    def triggered(self):
        self._action()


class XSectionMRUAction(Action):
    """A special action to implement the cross section MRU menu item"""

    def __init__(self, action):
        """
        Parameters
        ----------
        action : callable
        """
        self._action = action
        self._script = None
        # self.title = None
        # self.visible = False

    def triggered(self):
        self._action(self.script)

    @property
    def script(self):
        return self._script

    @script.setter
    def script(self, s):
        self._script = s
        self.visible = s is not None
        if s:
            self.title = os.path.basename(s)


class XSectionScriptEnvironment:
    """The cross section script environment"""

    def __init__(self, menu_name="pyxs"):
        self._menu_name = menu_name

        app = Application.instance()
        mw = app.main_window()
        if mw is None:
            print("none")
            return

        def _on_triggered_callback():
            """Load pyxs script menu action.

            Load new .pyxs file and run it.
            """
            view = Application.instance().main_window().current_view()
            if not view:
                MessageBox.critical(
                    "Error",
                    "No view open for creating the cross-" "section from",
                    MessageBox.b_ok(),
                )
                return None

            filename = FileDialog.get_open_file_name(
                "Select cross-section script",
                "",
                "XSection Scripts (*.pyxs);;All Files (*)",
            )

            # run the script and save it
            if filename.has_value():
                self.run_script(filename.value())
                self.make_mru(filename.value())

        def _XSectionMRUAction_callback(script):
            """*.pyxs menu action

            Load selected .pyxs file and run it.

            Parameters
            ----------
            script : str
            """
            self.run_script(script)
            self.make_mru(script)

        # Create pyxs submenu in Tools
        menu = mw.menu()

        if not menu.is_valid(f"tools_menu.{self._menu_name}_script_group"):
            menu.insert_separator("tools_menu.end", f"{self._menu_name}_script_group")
            menu.insert_menu(
                "tools_menu.end",
                f"{self._menu_name}_script_submenu",
                self._menu_name,
            )

        # Create Load XSection.py Script item in XSection (py)
        # global pyxs_script_load_menuhandler
        self.pyxs_script_load_menuhandler = MenuHandler(
            "Load pyxs script", _on_triggered_callback
        )
        menu.insert_item(
            f"tools_menu.{self._menu_name}_script_submenu.end",
            f"{self._menu_name}_script_load",
            self.pyxs_script_load_menuhandler,
        )
        menu.insert_separator(
            f"tools_menu.{self._menu_name}_script_submenu.end.end",
            f"{self._menu_name}_script_mru_group",
        )

        # Create list of existing pyxs scripts item in pyxs
        self._mru_actions = []
        for i in range(N_PYXS_SCRIPTS_MAX):
            a = XSectionMRUAction(_XSectionMRUAction_callback)
            self._mru_actions.append(a)
            menu.insert_item(
                f"tools_menu.{self._menu_name}_script_submenu.end",
                f"{self._menu_name}_script_mru{i}",
                a,
            )
            a.script = None

        # try to save the MRU list to $HOME/.klayout-pyxs-scripts
        i = 0
        home = os.getenv("HOME", None) or os.getenv("HOMESHARE", None)
        global pyxs_scripts
        if pyxs_scripts:
            for i, script in enumerate(pyxs_scripts.split(":")):
                if i < len(self._mru_actions):
                    self._mru_actions[i].script = script
        elif home:
            fn = os.path.join(home, ".klayout-pyxs-scripts")
            try:
                with open(fn) as file:
                    for line in file.readlines():
                        match = re.match(r"<mru>(.*)<\/mru>", line)
                        if match:
                            if i < len(self._mru_actions):
                                self._mru_actions[i].script = match.group(1)
                            i += 1
            except:
                pass

    def run_script(self, filename, p1=None, p2=None):
        """Run .pyxs script

        filename : str
            path to the .pyxs script
        """
        view = Application.instance().main_window().current_view()
        if not view:
            raise UserWarning("No view open for running the pyxs script")

        if p1 is None or p2 is None:
            app = Application.instance()
            scr_view = app.main_window().current_view()  # type: LayoutView
            scr_view_idx = app.main_window().current_view_index
            if not scr_view:
                MessageBox.critical(
                    "Error",
                    "No view open for creating the cross-" "section from",
                    MessageBox.b_ok(),
                )
                return False

            rulers = list(scr_view.each_annotation())

            if not rulers:
                MessageBox.critical(
                    "Error",
                    "No ruler present for the cross " "section line",
                    MessageBox.b_ok(),
                )
                return None

            p1_arr, p2_arr, ruler_text_arr = [], [], []

            for ruler in rulers:
                p1_arr.append(ruler.p1)
                p2_arr.append(ruler.p2)
                ruler_text_arr.append(ruler.text().split(".")[0])

        else:
            p1_arr, p2_arr, ruler_text_arr = [p1], [p2], [""]
            scr_view_idx = None

        target_views = []
        for p1_, p2_, text_ in zip(p1_arr, p2_arr, ruler_text_arr):

            if scr_view_idx:
                # return to the original view to run it again
                app.main_window().select_view(scr_view_idx)

            view = XSectionGenerator(filename).run(p1_, p2_, text_)
            target_views.append(view)

        return target_views
        # try:
        #     # print('XSectionGenerator(filename).run()')
        #     XSectionGenerator(filename).run()
        # except Exception as e:
        #     MessageBox.critical("Script failed", str(e),
        #                             MessageBox.b_ok())

    def make_mru(self, script):
        """Save list of scripts

        script : str
            path to the script to be saved
        """
        # Don't maintain MRU if an external list is provided
        global pyxs_scripts
        if pyxs_scripts:
            return

        # Make a new script list. New script goes first, ...
        scripts = [script]
        # ... the rest are taken from the existing list
        for a in self._mru_actions:
            if a.script != script:
                scripts.append(a.script)

        # make sure the list is filled to the same length
        while len(scripts) < len(self._mru_actions):
            scripts.append(None)

        # update list of actions
        for i in range(len(self._mru_actions)):
            self._mru_actions[i].script = scripts[i]

        if home := os.getenv("HOME", None) or os.getenv("HOMESHARE", None):
            fn = os.path.join(home, ".klayout-pyxs-scripts")
            with open(fn, "w") as file:
                file.write("<pyxs>\n")
                for a in self._mru_actions:
                    if a.script:
                        file.write(f"<mru>{a.script}</mru>\n")
                file.write("</pyxs>\n")


if __name__ == "__main__":
    import doctest

    doctest.testmod()



#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A feasibility study for a cross section generation using
# boolean operations. See "cmos.pyxs" for a brief description of the
# commands available and some examples.
# TODO: use a much smaller dbu for the simulation to have a really small delta
# the paths used for generating the masks are somewhat too thick
# TODO: the left and right areas are not treated correctly
import os
import re

from klayout_pyxs import (
    Action,
    Application,
    Box,
    FileDialog,
    LayerInfo,
    MessageBox,
    Point,
    Polygon,
)
from klayout_pyxs.compat import range, zip
from klayout_pyxs.geometry_2d import EP, LayoutData, parse_grow_etch_args
from klayout_pyxs.geometry_3d import LP, MaterialLayer, layer_to_tech_str, lp
from klayout_pyxs.layer_parameters import string_to_layer_info_params
from klayout_pyxs.utils import info, int_floor, make_iterable, print_info

# from importlib import reload
# try:
#     reload(klayout_pyxs.misc)
#     reload(klayout_pyxs.geometry_2d)
#     reload(klayout_pyxs.geometry_3d)
# except:
#     pass


info("Module pyxs3D_lib.py reloaded")
MIN_EXPORT_LAYER_THICKNESS = 5


class MaterialData3D:
    """Class to operate 3D materials.

    3D material is described by its top view (mask), vertical
    position (elevation), and thickness.

    """

    def __init__(self, layers, xs, delta):
        """
        Parameters
        ----------
        layers : list of geometry_3d.MaterialLayer

        xs: XSectionGenerator
        delta : float
            the intrinsic height (required for mask data because there
            cannot be an infinitely small mask layer (in database units)
        """
        self._layers = []
        self.data = layers
        self._delta = delta
        self._xs = xs
        self._lp = lp

    def __str__(self):
        n_layers = self.n_layers

        s = f"MaterialData3D (n_layers = {n_layers}, delta = {self._delta})"

        if n_layers > 0:
            s += ":"

        for li in range(min(5, n_layers)):
            s += f"\n    {str(self._layers[li])}"
        return s

    @property
    def data(self):
        """
        Return
        ------
        data: list of MaterialLayer
            layers which constitute the material
        """
        return self._layers

    @data.setter
    def data(self, layers):
        """
        Parameters
        ----------
        layers: list of MaterialLayer
            layers to be saved in the material
        """
        for la, lb in zip(layers[:-1], layers[1:]):
            if la.top > lb.bottom:
                raise ValueError(
                    f"layers must be a sorted list of non-overlapping layers. layers {la} and {lb} are not sorted and/or overlap."
                )

        self._layers = layers

    def add(self, other):
        """Add more material layers to the current one (OR).

        Parameters
        ----------
        other : MaterialData3D or list of MaterialLayer

        """
        other_layers = self._get_layers(other)
        self._layers = self._lp.boolean_l2l(self._layers, other_layers, LP.ModeOr)

    def and_(self, other):
        """Calculate overlap of the material with another material (AND).

        Parameters
        ----------
        other : MaterialData3D or list of MaterialLayer

        Returns
        -------
        res : MaterialData33D
        """
        other_layers = self._get_layers(other)
        return MaterialData3D(
            self._lp.boolean_l2l(self._layers, other_layers, LP.ModeAnd),
            self._xs,
            self._delta,
        )

    def inverted(self):
        """Calculate inversion of the material.

        Total region is determined by self._xs.background().

        Returns
        -------
        res : MaterialData3D
        """
        return MaterialData3D(
            self._lp.boolean_l2l(
                self._layers,
                [
                    MaterialLayer(
                        LayoutData([Polygon(self._xs.background())], self._xs),
                        -(self._xs.depth_dbu + self._xs.below_dbu),
                        self._xs.depth_dbu + self._xs.below_dbu + self._xs.height_dbu,
                    )
                ],
                EP.ModeXor,
            ),
            self._xs,
            self._delta,
        )

    def mask(self, other):
        """Mask material with another material (AND).

        Parameters
        ----------
        other : MaterialData3D or list of MaterialLayer

        """
        other_layers = self._get_layers(other)
        self._layers = self._lp.boolean_l2l(self._layers, other_layers, LP.ModeAnd)

    @property
    def n_layers(self):
        """
        Returns
        -------
        n_layers : int
            number of layers contained in the material

        """
        return len(self._layers)

    def not_(self, other):
        """Calculate difference with another material.

        Parameters
        ----------
        other : MaterialData3D or list of MaterialLayer

        Returns
        -------
        res : MaterialData3D
        """
        other_layers = self._get_layers(other)
        return MaterialData3D(
            self._lp.boolean_l2l(self._layers, other_layers, LP.ModeANotB),
            self._xs,
            self._delta,
        )

    def or_(self, other):
        """Calculate sum with another material (OR).

        Parameters
        ----------
        other : MaterialData3D or list of MaterialLayer

        Returns
        -------
        res : MaterialData3D
        """
        other_layers = self._get_layers(other)
        return MaterialData3D(
            self._lp.boolean_l2l(self._layers, other_layers, LP.ModeOr),
            self._xs,
            self._delta,
        )

    def sized(self, dx, dy=None):
        """Calculate material with a sized masks.

        Parameters
        ----------
        dx : float
            size change in x-direction in [um]
        dy : float (optional)
            size change in y-direction in [um]. Equals to dx by default.

        Returns
        -------
        res : MaterialData3D
        """
        dy = dx if dy is None else dy

        new_layers = []
        for l in self._layers:
            ld = l.mask.sized(dx, dy)
            new_layers.append(MaterialLayer(ld, l.bottom, l.thickness))

        return MaterialData3D(new_layers, self._xs, self._delta)

    def sub(self, other):
        """Subtract another material.

        Parameters
        ----------
        other : MaterialData3D or list of MaterialLayer

        """
        other_layers = self._get_layers(other)
        self._layers = self._lp.boolean_l2l(self._layers, other_layers, LP.ModeANotB)

    def transform(self, t):
        """Transform material masks with a transformation.

        Parameters
        ----------
        t : Trans
            transformation to be applied
        """
        for l in self._layers:
            l.mask.data = [p.transformed(t) for p in l.mask.data]

    def xor(self, other):
        """Calculate XOR with another material.

        Parameters
        ----------
        other : MaterialData3D or list of MaterialLayer

        Returns
        -------
        res : MaterialData3D
        """
        other_layers = self._get_layers(other)
        return MaterialData3D(
            self._lp.boolean_l2l(self._layers, other_layers, LP.ModeXor),
            self._xs,
            self._delta,
        )

    def close_gaps(self):
        """Close gaps in the polygons of the masks.

        Increase size of all polygons by 1 dbu in all directions.
        """
        sz = 1

        for l in self._layers:
            d = l.mask.data
            d = self._lp.size_p2p(d, 0, sz)
            d = self._lp.size_p2p(d, 0, -sz)
            d = self._lp.size_p2p(d, sz, 0)
            d = self._lp.size_p2p(d, -sz, 0)
            l.mask.data = d

    def remove_slivers(self):
        """Remove slivers in the polygons of the masks."""
        sz = 1
        for l in self._layers:
            d = l.mask.data
            d = self._lp.size_p2p(d, 0, -sz)
            d = self._lp.size_p2p(d, 0, sz)
            d = self._lp.size_p2p(d, -sz, 0)
            d = self._lp.size_p2p(d, sz, 0)
            l.mask.data = d

    def grow(
        self,
        z,
        xy=0.0,
        into=[],
        on=[],
        through=[],
        mode="square",
        buried=None,
        taper=None,
        bias=None,
    ):
        """
        Parameters
        ----------
        z : float
            grow height in [um]
        xy : float
            mask extension in [um]
        mode : str
            'round|square|octagon'. The profile mode.
        taper : float
            The taper angle. This option specifies tapered mode and cannot
            be combined with :mode.
        bias : float
            Adjusts the profile by shifting it to the interior of the figure.
            Positive values will reduce the line width by twice the value.
        on : list of MaterialData3D (optional)
            A material or an array of materials onto which the material is
            deposited (selective grow). The default is "all". This option
            cannot be combined with ":into". With ":into", ":through" has the
            same effect than ":on".
        into : list of MaterialData3D (optional)
            Specifies a material or an array of materials that the new
            material should consume instead of growing upwards. This will
            make "grow" a "conversion" process like an implant step.
        through : list of MaterialData3D (optional)
            To be used together with ":into". Specifies a material or an array
            of materials to be used for selecting grow. Grow will happen
            starting on the interface of that material with air, pass
            through the "through" material (hence the name) and consume and
            convert the ":into" material below.
        buried : float
            Applies the conversion of material at the given depth below the
            mask level. This is intended to be used together with :into
            and allows modeling of deep implants. The value is the depth
            below the surface.

        Returns
        -------
        res : MaterialData3D
            grown material
        """

        # parse the arguments
        into, through, on, mode = parse_grow_etch_args(
            "grow", MaterialData3D, into=into, on=on, through=through, mode=mode
        )

        # produce the geometry of the new material
        layers = self.produce_geom(
            "grow", xy, z, into, through, on, taper, bias, mode, buried
        )

        # prepare the result
        res = MaterialData3D(layers, self._xs, self._delta)

        # consume material
        if into:
            for i in into:  # i is MaterialData3D
                i.sub(res)
                i.remove_slivers()
        else:
            self._xs.air().sub(res)  # self._xs.air() is MaterialData3D
            self._xs.air().remove_slivers()
        return res

    def etch(
        self,
        z,
        xy=0.0,
        into=[],
        through=[],
        mode="square",
        taper=None,
        bias=None,
        buried=None,
    ):
        """

        Parameters
        ----------
        z : float
            etch depth in [um]
        xy : float (optional)
            mask extension, lateral in [um]
        mode : str
            'round|square|octagon'. The profile mode.
        taper :	float
            The taper angle. This option specifies tapered mode and cannot
            be combined with mode.
        bias : float
            Adjusts the profile by shifting it to the interior of the
            figure. Positive values will reduce the line width by twice
            the value.
        into :	list of MaterialData3D (optional)
            A material or an array of materials into which the etch is
            performed. This specification is mandatory.
        through : list of MaterialData3D (optional)
            A material or an array of materials which form the selective
            material of the etch. The etch will happen only where this
            material interfaces with air and pass through this material
            (hence the name).
        buried : float
            Applies the etching at the given depth below the surface. This
            option allows to create cavities. It specifies the vertical
            displacement of the etch seed and there may be more applications
            for this feature.

        """
        # parse the arguments
        into, through, on, mode = parse_grow_etch_args(
            "etch", MaterialData3D, into=into, through=through, on=None, mode=mode
        )

        if not into:
            raise ValueError("'etch' method: requires an 'into' specification")

        # prepare the result
        layers = self.produce_geom(
            "etch", xy, z, into, through, on, taper, bias, mode, buried
        )

        # produce the geometry of the etched material
        res = MaterialData3D(layers, self._xs, self._delta)

        # consume material and add to air
        if into:
            for i in into:  # i is MaterialData3D
                i.sub(res)
                i.remove_slivers()

        self._xs.air().add(res)  # self._xs.air() is MaterialData3D
        self._xs.air().close_gaps()

    @print_info(True)
    def produce_geom(self, method, xy, z, into, through, on, taper, bias, mode, buried):
        """

        method : str
        xy : float
            mask extension, lateral in [um]
        z : float
            vertical material size in [um]
        into : list of MaterialData3D
        through : list of MaterialData3D
        on : list of MaterialData3D
        taper : float
        bias : float
        mode : str
            'round|square|octagon'
        buried : float

        Returns
        -------
        layers : list of MaterialLayer
        """
        info(
            f"    method={method}, xy={xy}, z={z}, \n    into={into}, through={through}, on={on}, \n    taper={taper}, bias={bias}, mode={mode}, buried={buried})"
        )

        prebias = bias or 0.0

        if xy < 0.0:
            xy = -xy
            prebias += xy

        """
        if taper:
            d = z * math.tan(math.pi / 180.0 * taper)
            prebias += d - xy
            xy = d
        """
        # determine the "into" material by joining the layers of all "into"
        # materials or taking air's layers if required.
        # Finally we get a into_layers : list of MaterialLayer
        if into:
            into_layers = []
            for i in into:
                info(f"    i = {i}")
                if len(into_layers) == 0:
                    into_layers = i.data
                else:
                    into_layers = self._lp.boolean_l2l(i.data, into_layers, LP.ModeOr)
        else:
            # when deposit or grow is selected, into_layers is self._xs.air()
            into_layers = self._xs.air().data

        info(f"    into_layers = {into_layers}")

        # determine the "through" material by joining the layers of all
        # "through" materials
        # Finally we get a thru_layers : list of MaterialLayer
        if through:
            thru_layers = []
            for t in through:
                if len(thru_layers) == 0:
                    thru_layers = t.data
                else:
                    thru_layers = self._lp.boolean_l2l(t.data, thru_layers, LP.ModeOr)
            info(f"    thru_layers = {thru_layers}")

        # determine the "on" material by joining the data of all "on" materials
        # Finally we get an on_layers : list of MaterialLayer
        if on:
            on_layers = []
            for o in on:
                if len(on_layers) == 0:
                    on_layers = o.data
                else:
                    on_layers = self._lp.boolean_l2l(o.data, on_layers, LP.ModeOr)
            info(f"    on_layers = {on_layers}")

        offset = self._delta
        layers = self._layers
        info(f"    Seed material to be grown: {self}")

        """
        if abs(buried or 0.0) > 1e-6:
            t = Trans(Point(
                    0, -_int_floor(buried / self._xs.dbu + 0.5)))
            d = [p.transformed(t) for p in d]
        """

        # in the "into" case determine the interface region between
        # self and into
        if into or through or on:
            # apply an artificial sizing to create an overlap before
            if offset == 0:
                offset = self._xs.delta_dbu / 2
                layers = self._lp.size_l2l(layers, 0, dz=offset)

            if on:
                layers = self._lp.boolean_l2l(layers, on_layers, EP.ModeAnd)
            elif through:
                layers = self._lp.boolean_l2l(layers, thru_layers, EP.ModeAnd)
            else:
                layers = self._lp.boolean_l2l(layers, into_layers, EP.ModeAnd)
        info(f"    overlap layers = {layers}")

        pi = int_floor(prebias / self._xs.dbu + 0.5)
        info(f"    pi = {pi}")
        if pi < 0:
            layers = self._lp.size_l2l(layers, -pi, dy=-pi, dz=0)
        elif pi > 0:
            raise NotImplementedError("pi > 0 not implemented yet")
        xyi = int_floor(xy / self._xs.dbu + 0.5)  # size change in [dbu]
        zi = int_floor(z / self._xs.dbu + 0.5) - offset  # height in [dbu]
        info(f"    xyi = {xyi}, zi = {zi}")

        if taper:
            raise NotImplementedError("taper option is not supported yet")
            # d = self._ep.size_p2p(d, xyi, zi, 0)
        elif xyi <= 0:
            layers = self._lp.size_l2l(layers, 0, dy=0, dz=zi)
            # d = self._ep.size_p2p(d, 0, zi)
        elif mode in ["round", "square"]:
            # same as square for now
            layers = self._lp.size_l2l(layers, xyi, dy=xyi, dz=zi)

            # raise NotImplementedError('round option is not supported yet')
            # emulate "rounding" of corners by performing soft-edged sizes
            # d = self._ep.size_p2p(d, xyi / 3, zi / 3, 1)
            # d = self._ep.size_p2p(d, xyi / 3, zi / 3, 0)
            # d = self._ep.size_p2p(d, xyi - 2 * (xyi / 3), zi - 2 * (zi / 3), 0)
        elif mode == "octagon":
            raise NotImplementedError("octagon option is not supported yet")
            # d = self._ep.size_p2p(d, xyi, zi, 1)

        if through:
            layers = self._lp.boolean_l2l(layers, thru_layers, LP.ModeANotB)

        info("    layers before and with into:")
        layers = self._lp.boolean_l2l(layers, into_layers, LP.ModeAnd)
        info("    layers after and with into:")

        info(f"    final layers = {layers}")
        return layers

    @staticmethod
    def _get_layers(m):
        if isinstance(m, MaterialData3D):
            return m.data
        elif isinstance(m, (tuple, list)):
            return m
        else:
            raise TypeError(
                f"m should be either an instance of MaterialData3D or a list of MaterialLayer. {type(m)} is given."
            )


class XSectionGenerator:
    """The main class that creates a cross-section file

    Attributes
    ----------
    _lp : LayerProcessor

    """

    def __init__(self, file_path):
        """
        Parameters
        ----------
        file_path : str
        """
        # TODO: adjust this path:
        self._file_path = file_path
        self._lyp_file = None
        self._lp = lp
        self._flipped = False
        self._box_dbu = None  # pya.Box
        self._air, self._air_below = None, None
        self._delta, self._extend = 1, None
        self._height, self._depth, self._below = None, None, None
        self._thickness_scale_factor = 1
        self._target_gds_file_name = ""
        self._target_tech_file_name = ""
        self._tech_str = ""

        self.set_output_parameters(filename="3d_xs.gds")

    def layer(self, layer_spec):
        """Fetches an input layer from the original layout.

        Parameters
        ----------
        layer_spec : str

        Returns
        -------
        ld : LayerData
        """
        ld = LayoutData([], self)  # empty

        # collect shapes from the corresponding layer touching
        # extended self._box_dbu into ld._polygons
        ld.load(
            self._layout,
            self._cell,
            self._box_dbu.enlarged(Point(self._extend, self._extend)),
            layer_spec,
        )

        return ld

    @print_info(True)
    def mask(self, layer_data):
        """Designates the layout_data object as a litho pattern (mask).

        This is the starting point for structured grow or etch operations.

        Parameters
        ----------
        layer_data : LayoutData

        Returns
        -------
        MaskData
        """
        info(f"    layer_data = {layer_data}")
        mask = layer_data.and_([Polygon(self._box_dbu)])
        info(f"    mask = {mask}")
        return self._mask_to_seed_material(mask)

    # @property
    def air(self):
        """Return a material describing the air above

        Return
        ------
        air : MaterialData3D
        """
        return self._air

    # @property
    def bulk(self):
        """Return a material describing the wafer body

        Return
        ------
        bulk : MaterialData3D
        """
        return MaterialData3D(self._bulk.data, self, self._delta)

    @print_info(True)
    def output(self, layer_spec, material, color=None):
        """Outputs a material object to the output layout

        Parameters
        ----------
        layer_spec : str
            layer specification
        material : MaterialData3D
        """
        if not isinstance(material, MaterialData3D):
            raise TypeError(
                f"'output' method: second parameter must be a material object (MaterialData3D). {type(material)} is given"
            )

        # confine the shapes to the region of interest
        # info('    roi = {}'.format(self._roi))
        # info('    material = {}'.format(material))
        export_layers = self._lp.boolean_l2l(self._roi, material.data, LP.ModeAnd)
        info(f"    layers to export = {export_layers}")
        l, data_type, name = string_to_layer_info_params(layer_spec, True)
        # info('{}, {}, {}'.format(l, data_type, name))

        name = name or ""

        for i, layer in enumerate(export_layers):
            layer_not_empty = False
            layer_no = l + i
            if layer.thickness < MIN_EXPORT_LAYER_THICKNESS:
                continue  # next layer in the material

            ls = LayerInfo(layer_no, data_type, f"{name} ({layer.bottom}-{layer.top})")
            li = self._target_layout.insert_layer(ls)
            shapes = self._target_layout.cell(self._target_cell).shapes(li)
            for polygon in layer.mask.data:
                # info('S = {}, S_box = {}'
                #      .format(polygon.area(), polygon.bbox().area()))
                if polygon.area() > 0.001 * polygon.bbox().area():
                    layer_not_empty = True
                    shapes.insert(polygon)

            if layer_not_empty:
                self._tech_str += layer_to_tech_str(
                    layer_no, layer, name=name, color=color
                )

        # info('    {}'.format(self._tech_str[0:150]))

    @print_info(True)
    def all(self):
        """A pseudo-mask, covering the whole wafer

        Return
        ------
        res : MaterialData3D
        """
        res = self._mask_to_seed_material(LayoutData([Polygon(self._box_dbu)], self))
        info(f"    result: {res}")
        return res

    def flip(self):
        """Start or end backside processing"""
        self._air, self._air_below = self._air_below, self._air
        self._flipped = not self._flipped

    def diffuse(self, *args, **kwargs):
        """Same as deposit()"""
        return self.all().grow(*args, **kwargs)

    def deposit(self, *args, **kwargs):
        """Deposits material as a uniform sheet.

        Equivalent to all.grow(...)

        Return
        ------
        res : MaterialData3D
        """
        return self.all().grow(*args, **kwargs)

    def grow(self, *args, **kwargs):
        """Same as deposit()"""
        return self.all().grow(*args, **kwargs)

    def etch(self, *args, **kwargs):
        """Uniform etching

        Equivalent to all.etch(...)

        """
        return self.all().etch(*args, **kwargs)

    def planarize(self, into=[], downto=[], less=None, to=None, **kwargs):
        """Planarization"""

        if not into:
            raise ValueError("'planarize' requires an 'into' argument")

        into = make_iterable(into)
        for i in into:
            # should be MaterialData @@@
            if not isinstance(i, MaterialData3D):
                raise TypeError(
                    "'planarize' method: 'into' expects "
                    "a material parameter or an array "
                    "of such"
                )

        downto = make_iterable(downto)
        for i in downto:
            # should be MaterialData @@@
            if not isinstance(i, MaterialData3D):
                raise TypeError(
                    "'planarize' method: 'downto' expects "
                    "a material parameter or an array "
                    "of such"
                )

        if less is not None:
            less = int_floor(0.5 + float(less) / self.dbu)

        if to is not None:
            to = int_floor(0.5 + float(to) / self.dbu)

        if downto:
            downto_data = []
            for d in downto:
                if len(downto_data) == 0:
                    downto_data = d.data
                else:
                    downto_data = self._lp.boolean_p2p(d.data, downto_data, LP.ModeOr)

            # determine upper bound of material
            if downto_data:
                raise NotImplementedError("downto not implemented yet")
        elif into and not to:
            raise NotImplementedError("into and not to not implemented yet")
        if to:
            less = less or 0
            if self._flipped:
                removed_box = MaterialLayer(
                    LayoutData(
                        [
                            Polygon(
                                self._box_dbu.enlarged(
                                    Point(self._extend, self._extend)
                                )
                            )
                        ],
                        self,
                    ),
                    -(self.depth_dbu + self.below_dbu),
                    (to + less) + self.depth_dbu + self.below_dbu,
                )

            else:
                removed_box = MaterialLayer(
                    LayoutData(
                        [
                            Polygon(
                                self._box_dbu.enlarged(
                                    Point(self._extend, self._extend)
                                )
                            )
                        ],
                        self,
                    ),
                    to - less,
                    self.height_dbu - (to - less),
                )

            rem = MaterialData3D([], self, self._delta)
            for i in into:
                rem.add(i.and_([removed_box]))
                i.sub([removed_box])

            self.air().add(rem)
            self.air().close_gaps()

    def set_thickness_scale_factor(self, factor):
        """Configures layer thickness scale factor to have better proportions

        Parameters
        ----------
        factor : float
        """
        self._thickness_scale_factor = factor

    def set_output_parameters(self, filename=None, format=None):
        if filename:
            self._target_gds_file_name = filename
            self._target_tech_file_name = f"{filename}_tech"

    def set_delta(self, x):
        """Configures the accuracy parameter"""
        self._delta = int_floor(x / self._dbu + 0.5)
        info(f"XSG._delta set to {self._delta}")

    @property
    def delta_dbu(self):
        return self._delta

    def set_height(self, x):
        """Configures the height of the processing window

        Parameters
        ----------
        x : float
            height in [um]

        """
        self._height = int_floor(x / self._dbu + 0.5)
        info(f"XSG._height set to {self._height}")
        self._update_basic_regions()

    @property
    def height_dbu(self):
        return self._height

    def set_depth(self, x):
        """Configures the depth of the processing window
        or the wafer thickness for backside processing (see `below`)

        Parameters
        ----------
        x : float
            depth of the wafer in [um]
        """
        self._depth = int_floor(x / self._dbu + 0.5)
        info(f"XSG._depth set to {self._depth}")
        self._update_basic_regions()

    @property
    def depth_dbu(self):
        return self._depth

    def set_below(self, x):
        """Configures the lower height of the processing window for backside processing

        Parameters
        ----------
        x : float
            depth below the wafer in [um]

        """
        self._below = int_floor(x / self._dbu + 0.5)
        info(f"XSG._below set to {self._below}")
        self._update_basic_regions()

    @property
    def below_dbu(self):
        return self._below

    def set_extend(self, x):
        """Set the extend of the computation region

        Parameters
        ----------
        x : float
            extend in [um]
        """
        self._extend = int_floor(x / self._dbu + 0.5)
        self._update_basic_regions()

    @property
    def extend_dbu(self):
        return self._extend

    @property
    def width_dbu(self):
        """Box size along the major flat (x-direction).

        Returns
        -------
        width : int
        """
        return self._box_dbu.width()

    @property
    def breadth_dbu(self):
        """Box size perpendicular to major flat (y-direction)

        Returns
        -------
        breadth : int
        """
        return self._box_dbu.height()

    def background(self):
        """This is used in inverted() method.

        Returns
        -------
        res : Box
            The extended box including the ruler.
        """
        # x1, y1 = self._line_dbu.p1.x, self._line_dbu.p1.y
        # x2, y2 = self._line_dbu.p2.x, self._line_dbu.p2.y
        # if x2 < x1:
        #     x1, x2 = x2, x1

        # if y2 < y1:
        #     y1, y2 = y2, y1

        # x1 -= self._extend
        # y1 -= self._extend
        # x2 += self._extend
        # y2 += self._extend
        # return pya.Box(pya.Point(x1 - self._delta * 5, y1 - self._delta * 5),
        #                pya.Point(x2 + self._delta * 5, y2 + self._delta * 5))

        # Since we will be inverting only masks, we can simply return a box
        # to be exported to 3D.
        return self._box_dbu  # TODO: return an extended box

    @property
    def dbu(self):
        """
        Returns
        -------
        dbu : float
            Database unit
        """
        return self._dbu

    def layers_file(self, lyp_file):
        """Configures a .lyp layer properties file to be used on the
        cross-section layout

        Parameters
        ----------
        lyp_file : str
            path to the lyp file
        """
        self._lyp_file = lyp_file

    @print_info(False)
    def run(self):
        """The basic generation method"""

        if not self._setup():
            return None

        self._update_basic_regions()

        text = None
        with open(self._file_path) as file:
            text = file.read()

        if not text:
            MessageBox.critical(
                "Error", "Error reading file #{self._file_path}", MessageBox.b_ok()
            )
            return None

        # prepare variables to be visible in the script
        locals_ = dir(self)
        locals_dict = {attr: getattr(self, attr) for attr in locals_ if attr[0] != "_"}
        try:
            exec(text, locals_dict)
        except Exception as e:
            # For development
            # print(e.__traceback__.)
            # print(dir(e))
            MessageBox.critical("Error", str(e), MessageBox.b_ok())
            # pass
            return None

        Application.instance().main_window().cm_lv_add_missing()  # @@@
        if self._lyp_file:
            self._target_view.load_layer_props(self._lyp_file)
        self._target_view.zoom_fit()
        self._target_layout.write(self._target_gds_file_name)

        info(f"    len(bulk.data) = {len(self._bulk.data)}")
        self._tech_str = (
            "# This file was generated automatically by pyxs.\n\n"
            + layer_to_tech_str(255, self._bulk.data[0], "Substrate")
            + self._tech_str
        )
        with open(self._target_tech_file_name, "w") as f:
            f.write(self._tech_str)

        return None

    @print_info(True)
    def _mask_to_seed_material(self, mask):
        """Convert mask to a seed material for growth / etch operations.

        Parameters
        ----------
        mask: LayoutData
            top view of the region to be grown / etched

        Return
        ------
        seed : MaterialData3D
            Thin seed material to be used in geometry generation.
        """
        info(f"    mask = {mask}")

        mask_material = [
            MaterialLayer(
                mask,
                -(self._depth + self._below),
                self._depth + self._below + self._height,
            )
        ]
        info(f"    mask material = {mask}")

        air = self._air.data
        info(f"    air =        {air}")

        air_sized = self._lp.size_l2l(air, 0, 0, self._delta)
        info(f"    air sized =  {air_sized}")

        # extended air minus air
        air_border = self._lp.boolean_l2l(air_sized, air, LP.ModeANotB)
        info(f"    air_border = {air_border}")

        # overlap of air border and mask layer
        seed_layers = self._lp.boolean_l2l(air_border, mask_material, EP.ModeAnd)

        info(f"    seed_layers= {seed_layers}")

        return MaterialData3D(seed_layers, self, self._delta)

    @print_info(False)
    def _update_basic_regions(self):

        h = self._height  # height above the wafer
        d = self._depth  # thickness of the wafer
        b = self._below  # distance below the wafer

        # w = self._line_dbu.length()  # length of the ruler
        e = self._extend  # extend to the sides

        # TODO: add extend to the basic regions
        self._area = [
            MaterialLayer(
                LayoutData([Polygon(self._box_dbu)], self), -(b + d), (b + d + h)
            )
        ]  # Box(-e, -(d+b), w+e, h)
        self._roi = [
            MaterialLayer(
                LayoutData([Polygon(self._box_dbu)], self), -(b + d), (b + d + h)
            )
        ]  # Box(0, -(d + b), w, h)

        self._air = MaterialData3D(
            [MaterialLayer(LayoutData([Polygon(self._box_dbu)], self), 0, h)], self, 0
        )
        self._air_below = MaterialData3D(
            [MaterialLayer(LayoutData([Polygon(self._box_dbu)], self), -(d + b), b)],
            self,
            0,
        )

        self._bulk = MaterialData3D(
            [MaterialLayer(LayoutData([Polygon(self._box_dbu)], self), -d, d)], self, 0
        )

        info(f"    XSG._area:      {self._area}")
        info(f"    XSG._roi:       {self._roi}")
        info(f"    XSG._air:       {self._air}")
        info(f"    XSG._bulk:      {self._bulk}")
        info(f"    XSG._air_below: {self._air_below}")

    @print_info(True)
    def _setup(self):

        # locate the layout
        app = Application.instance()
        view = app.main_window().current_view()  # LayoutView
        if not view:
            MessageBox.critical(
                "Error",
                "No view open for creating the cross " "section from",
                MessageBox.b_ok(),
            )
            return False

        # locate the (single) ruler
        rulers = []
        n_rulers = 0
        for a in view.each_annotation():
            # Use only rulers with "plain line" style
            # print(a.style)
            # print(Annotation.StyleLine)
            # if a.style == Annotation.StyleLine:
            rulers.append(a)
            n_rulers += 1

        # if n_rulers == 0 or n_rulers >= 2:
        #     MessageBox.info("No rulers",
        #                         "Number of rulers is not equal to one. "
        #                         "Will be exporting the whole layout",
        #                         pya.MessageBox.b_ok())

        # if n_rulers == 1:
        #     MessageBox.info(
        #             "Box export", "One ruler is present for the cross "
        #             "section line. Will be exporting only shapes in the box",
        #             pya.MessageBox.b_ok())

        cv = view.cellview(view.active_cellview_index())  # CellView
        if not cv.is_valid():
            MessageBox.critical(
                "Error", "The selected layout is not valid", MessageBox.b_ok()
            )
            return False

        self._cv = cv  # CellView
        self._layout = cv.layout()  # Layout
        self._dbu = self._layout.dbu
        self._cell = cv.cell_index  # int

        if n_rulers == 1:
            # get the start and end points in database units and micron
            p1_dbu = Point.from_dpoint(rulers[0].p1 * (1.0 / self._dbu))
            p2_dbu = Point.from_dpoint(rulers[0].p2 * (1.0 / self._dbu))
        else:
            # TODO: choose current cell, not top cell
            top_cell = self._layout.top_cell()
            p1_dbu = (top_cell.bbox().p1 * (1.0 / self._dbu)).dup()
            p1_dbu = top_cell.bbox().p1.dup()
            p2_dbu = (top_cell.bbox().p2 * (1.0 / self._dbu)).dup()
            p2_dbu = top_cell.bbox().p2.dup()
        self._box_dbu = Box(p1_dbu, p2_dbu)  # box describing the ruler
        info(f"XSG._box_dbu to be used is: {self._box_dbu}")

        # create a new layout for the output
        cv = app.main_window().create_layout(1)
        cell = cv.layout().add_cell("XSECTION")
        self._target_view = app.main_window().current_view()
        self._target_view.select_cell(cell, 0)
        self._target_layout = cv.layout()
        self._target_layout.dbu = self._dbu
        self._target_cell = cell

        # initialize height and depth
        self._extend = int_floor(2.0 / self._dbu + 0.5)  # 2 um in dbu
        self._delta = 10
        self._height = int_floor(2.0 / self._dbu + 0.5)  # 2 um in dbu
        self._depth = int_floor(2.0 / self._dbu + 0.5)  # 2 um in dbu
        self._below = int_floor(2.0 / self._dbu + 0.5)  # 2 um in dbu

        info(f"    XSG._dbu is:    {self._dbu}")
        info(f"    XSG._extend is: {self._extend}")
        info(f"    XSG._delta is:  {self._delta}")
        info(f"    XSG._height is: {self._height}")
        info(f"    XSG._depth is:  {self._depth}")
        info(f"    XSG._below is:  {self._below}")

        return True


# MENU AND ACTIONS
# ----------------
N_PYXS_SCRIPTS_MAX = 4

pyxs_script_load_menuhandler = None
pyxs_scripts = None


class MenuHandler(Action):
    """Handler for the load .xs file action"""

    def __init__(self, title, action, shortcut=None, icon=None):
        """
        Parameters
        ----------
        title : str
        action : callable
        shortcut : str
        icon : str
        """
        self.title = title
        self._action = action
        if shortcut:
            self.shortcut = shortcut
        if icon:
            self.icon = icon

    def triggered(self):
        self._action()


class XSectionMRUAction(Action):
    """A special action to implement the cross section MRU menu item"""

    def __init__(self, action):
        """
        Parameters
        ----------
        action : callable
        """
        self._action = action
        self._script = None

    def triggered(self):
        self._action(self.script)

    @property
    def script(self):
        return self._script

    @script.setter
    def script(self, s):
        self._script = s
        self.visible = s is not None
        if s:
            self.title = os.path.basename(s)


class XSectionScriptEnvironment:
    """The cross section script environment"""

    def __init__(self):
        app = Application.instance()
        mw = app.main_window()

        def _on_triggered_callback():
            """Load pyxs script menu action.

            Load new .pyxs file and run it.
            """
            view = Application.instance().main_window().current_view()
            if not view:
                raise UserWarning("No view open for running the pyxs script")

            filename = FileDialog.get_open_file_name(
                "Select cross-section script",
                "",
                "XSection Scripts (*.pyxs);;All Files (*)",
            )

            # run the script and save it
            if filename.has_value():
                self.run_script(filename.value())
                self.make_mru(filename.value())

        def _XSectionMRUAction_callback(script):
            """*.pyxs menu action

            Load selected .pyxs file and run it.

            Parameters
            ----------
            script : str
            """
            self.run_script(script)
            self.make_mru(script)

        # Create pyxs submenu in Tools
        menu = mw.menu()
        if not menu.is_valid("tools_menu.pyxs3D_script_group"):
            menu.insert_separator("tools_menu.end", "pyxs3D_script_group")
            menu.insert_menu("tools_menu.end", "pyxs3D_script_submenu", "pyxs3D")

        # Create Load XSectionpy Script item in XSection (py)
        global pyxs_script_load_menuhandler
        pyxs_script_load_menuhandler = MenuHandler(
            "Load pyxs script", _on_triggered_callback
        )
        menu.insert_item(
            "tools_menu.pyxs3D_script_submenu.end",
            "pyxs3D_script_load",
            pyxs_script_load_menuhandler,
        )
        menu.insert_separator(
            "tools_menu.pyxs3D_script_submenu.end.end", "pyxs3D_script_mru_group"
        )

        # Create list of existing pyxs scripts item in pyxs
        self._mru_actions = []
        for i in range(N_PYXS_SCRIPTS_MAX):
            a = XSectionMRUAction(_XSectionMRUAction_callback)
            self._mru_actions.append(a)
            menu.insert_item(
                "tools_menu.pyxs3D_script_submenu.end",
                f"pyxs3D_script_mru{i}",
                a,
            )
            a.script = None

        # try to save the MRU list to $HOME/.klayout-processing-mru
        i = 0
        home = os.getenv("HOME", None) or os.getenv("HOMESHARE", None)
        global pyxs_scripts
        if pyxs_scripts:
            for i, script in enumerate(pyxs_scripts.split(":")):
                if i < len(self._mru_actions):
                    self._mru_actions[i].script = script
        elif home:
            fn = home + "\\.klayout-pyxs-scripts"
            try:
                with open(fn) as file:
                    for line in file.readlines():
                        match = re.match(r"<mru>(.*)<\/mru>", line)
                        if match:
                            if i < len(self._mru_actions):
                                self._mru_actions[i].script = match.group(1)
                            i += 1
            except:
                pass

    def run_script(self, filename):
        """Run .pyxs script

        filename : str
            path to the .pyxs script
        """
        view = Application.instance().main_window().current_view()
        if not view:
            raise UserWarning("No view open for running the pyxs script")

        # cv = view.cellview(view.active_cellview_index())

        XSectionGenerator(filename).run()
        # try:
        #     # print('XSectionGenerator(filename).run()')
        #     XSectionGenerator(filename).run()
        # except Exception as e:
        #     MessageBox.critical("Script failed", str(e),
        #                             MessageBox.b_ok())

    def make_mru(self, script):
        """Save list of scripts

        script : str
            path to the script to be saved
        """
        # Don't maintain MRU if an external list is provided
        global pyxs_scripts
        if pyxs_scripts:
            return

        # Make a new script list. New script goes first, ...
        scripts = [script]
        # ... the rest are taken from the existing list
        for a in self._mru_actions:
            if a.script != script:
                scripts.append(a.script)

        # make sure the list is filled to the same length
        while len(scripts) < len(self._mru_actions):
            scripts.append(None)

        # update list of actions
        for i in range(len(self._mru_actions)):
            self._mru_actions[i].script = scripts[i]

        # try to save the MRU list to $HOME/.klayout-xsection
        home = os.getenv("HOME", None) or os.getenv("HOMESHARE", None)
        if home:
            fn = home + "\\.klayout-pyxs-scripts"
            with open(fn, "w") as file:
                file.write("<pyxs>\n")
                for a in self._mru_actions:
                    if a.script:
                        file.write(f"<mru>{a.script}</mru>\n")
                file.write("</pyxs>\n")


if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""klayout_pyxs.__init__.py

Copyright 2017-2019 Dima Pustakhod

"""
from ._version import __version__

DEBUG = False
HAS_KLAYOUT = False
HAS_PYA = False

try:
    if DEBUG:
        print("Trying to import klayout module... ", end="")
    import importlib

    importlib.import_module("klayout")
    HAS_KLAYOUT = True

    from klayout.db import Box, DPoint, Edge
    from klayout.db import EdgeProcessor as EP_
    from klayout.db import (
        Edges,
        LayerInfo,
        Point,
        Polygon,
        Region,
        SimplePolygon,
        Trans,
    )

    if DEBUG:
        print("found!")
except:
    if DEBUG:
        print("not found!")
    try:
        if DEBUG:
            print("Trying to import pya module... ", end="")
        import pya as klayout

        # For plugin only
        from pya import Action, Application, Box, DPoint, Edge
        from pya import EdgeProcessor as EP_
        from pya import (
            Edges,
            FileDialog,
            LayerInfo,
            MessageBox,
            Point,
            Polygon,
            Region,
            SimplePolygon,
            Trans,
        )

        HAS_PYA = True
        if DEBUG:
            print("found!")
    except:
        if DEBUG:
            print("not found!")
        raise ModuleNotFoundError(
            "Neither pya nor klayout module are not "
            "installed in the current python distribution."
        )


# from .misc import info
# reload(pyxs.misc)
# reload(pyxs.geometry_2d)
# reload(pyxs.geometry_3d)


def _poly_repr(self):
    """Return nice representation of the Polygon instance

    This is useful when printing a list of Polygons
    """
    return f"{self.num_points()} pts: {self.__str__()}"


Polygon.__repr__ = _poly_repr

# info('pyxs.__init__.py loaded')

from klayout_pyxs.pyxs_lib import XSectionScriptEnvironment

__version__ = "0.1.8"

__all__ = [
    "XSectionScriptEnvironment",
    "__version__",
]



""" pyxs.geometry_2d.py

(C) 2017 Dima Pustakhod and contributors
"""
from random import random

from klayout_pyxs.compat import range, zip
from klayout_pyxs.geometry_2d import EdgeProcessor, LayoutData
from klayout_pyxs.utils import info, print_info


class LayerProcessor(EdgeProcessor):
    """Class implementing operations on MaterialLayer lists"""

    def normalize(self, layers):
        """
        Parameters
        ----------
        layers : list of MaterialLayer or empty list
            a list of non-sorted and / or overlapping layers

        Returns
        -------
        res : list of MaterialLayer
            a sorted list of non-overlapping layers
        """
        layers.sort()
        res = self.split_overlapping_z(layers)
        res = self.merge_layers_same_mask(res)

        return res

    @print_info(False)
    def split_overlapping_z(self, layers):
        """
        Parameters
        ----------
        layers : list of MaterialLayer
            a list of non-sorted and / or overlapping layers

        Returns
        -------
        res : list of MaterialLayer
            a sorted list of non-overlapping layers
        """
        info(f"    layers = {layers}")
        _check_layer_list_sorted(layers)

        res = []

        while layers:
            la = layers.pop(0)  # first element is the lowest in z
            info(f"    la = {la}")
            if not layers:  # la was the only element
                res += [la]
                continue

            lb = layers.pop(0)  # take next element
            info(f"    lb = {lb}")
            if la.is_lower(lb, levela="top", levelb="bottom"):
                # a is lower or touching
                # layers is sorted, la will not overlap with other lb
                res += [la]
                layers.insert(0, lb)
                info("    la top < lb btm, la is moved to result, lb is returned")
            elif la.bottom == lb.bottom:
                info("    la btm == lb btm")
                if la.top < lb.top:
                    lb_split = lb.split_by_layer(la)
                    o = MaterialLayer(
                        la.mask.or_(lb_split[0].mask), la.bottom, la.thickness
                    )
                    layers = [o] + layers

                    # top part of b is inserted to layers, ensuring sorted order
                    i = 0
                    while i < len(layers):
                        if lb_split[1].bottom < layers[i].bottom:
                            layers.insert(i, lb_split[1])
                            break
                        elif lb_split[1].bottom == layers[i].bottom:
                            if lb_split[1].top <= layers[i].top:
                                layers.insert(i, lb_split[1])
                                break
                        i += 1
                    else:
                        layers.append(lb_split[1])
                    info("    la top < lb top, o calculated, added to layers")
                else:
                    # la is the same height as lb
                    # perform OR operation on the LayoutData
                    o = MaterialLayer(la.mask.or_(lb.mask), la.bottom, la.thickness)
                    layers.insert(0, o)
                    info("    la top == lb top, o calculated, added to layers")
            else:
                info(
                    "    la btm < lb btm, lower part of la is result, rest added to layers"
                )
                # lb bottom splits a somewhere (maybe lb top too)
                la_split = la.split_by_layer(lb)

                # bottom sublayer is not overlapping with lb and others
                res.append(la_split[0])
                layers = la_split[1:] + [lb] + layers

        info(f"    res = {res}")
        return res

    @print_info(False)
    def boolean_l2l(self, la, lb, mode, rh=True, mc=True):
        """
        Parameters
        ----------
        la : list of MaterialLayer or empty list
            sorted list. layers must not overlap with each other
        lb : list of MaterialLayer or empty list
            sorted list. layers must not overlap with each other
        mode: int
        rh : bool (optional)
            resolve_holes
        mc : bool (optional)
            min_coherence

        Returns
        -------
        list of MaterialLayer or []
        """
        n_la, n_lb = len(la), len(lb)  # number of polygons in pa and pb

        info(f"    n_la = {n_la}, n_lb = {n_lb}, mode = {mode}")

        ia, ib = 0, 0
        a = la[ia] if la else None
        b = lb[ib] if lb else None
        la_res, lb_res, oa, ob = [], [], [], []
        while a and b:
            info(f"    a = {a}")
            info(f"    b = {b}")
            if a.is_lower_s(b, "bottom"):
                info("    a bottom is lower")
                top = min(a.top, b.bottom)
                info(f"    top = {top}")
                if top == a.top:  # no overlap
                    info("    a top is lower than b bottom, no overlap")
                    la_res += [a]
                    ia += 1
                    a = None if ia >= len(la) else la[ia]
                else:
                    info("    a top is higher than b bottom, overlap")
                    # use part of a from a.bottom to top
                    la_res += [MaterialLayer(a.mask, a.bottom, top - a.bottom)]
                    # overlapping candidate a is a from top to a.top
                    a = MaterialLayer(a.mask, top, a.top - top)
                continue
            elif b.is_lower_s(a, "bottom"):
                info("    b is lower")
                top = min(b.top, a.bottom)
                if top == b.top:  # no overlap
                    lb_res += [b]
                    ib += 1
                    b = None if ib >= len(lb) else lb[ib]
                else:
                    # use part of b from b.bottom to top
                    lb_res += [MaterialLayer(b.mask, b.bottom, top - b.bottom)]
                    # overlapping candidate b is b from top to b.top
                    b = MaterialLayer(b.mask, top, b.top - top)
                continue
            else:
                assert a.bottom == b.bottom, "bottoms must be equal here"
                info("    same bottom")
                if a.is_lower_s(b, "top") or b.is_lower_s(a, "top"):
                    top = min(a.top, b.top)
                    if top < b.top:  # a is in the overlap, b is higher
                        info("    b is higher")
                        oa += [a]
                        ob += [MaterialLayer(b.mask, b.bottom, top - b.bottom)]
                        b = MaterialLayer(b.mask, top, b.top - top)  # remaining top
                        ia += 1
                        a = None if ia >= len(la) else la[ia]
                        continue
                    elif top < a.top:  # b is in the overlap, a is higher
                        info("    a is higher")
                        ob += [b]
                        oa += [MaterialLayer(a.mask, a.bottom, top - a.bottom)]
                        a = MaterialLayer(a.mask, top, a.top - top)  # remaining top
                        ib += 1
                        b = None if ib >= len(lb) else lb[ib]
                        continue
                else:
                    assert a.top == b.top, "tops must be equal here"
                    info("    same top")
                    oa += [a]
                    ob += [b]
                    ia += 1
                    a = None if ia >= len(la) else la[ia]
                    ib += 1
                    b = None if ib >= len(lb) else lb[ib]
                    continue

        if a:
            la_res += [a]
        if b:
            lb_res += [b]

        # add remaining a's and b's
        while ia < len(la) - 1:
            ia += 1
            la_res += [la[ia]]

        while ib < len(lb) - 1:
            ib += 1
            lb_res += [lb[ib]]

        lo_res = []
        for a, b in zip(oa, ob):
            if o_polygons := self.boolean_p2p(a.mask.data, b.mask.data, mode, rh, mc):
                lo_res += [
                    MaterialLayer(
                        LayoutData(o_polygons, a.mask._xs), a.bottom, a.top - a.bottom
                    )
                ]

        info(f"    la_res = {la_res}")
        info(f"    lb_res = {lb_res}")
        info(f"    lo_res = {lo_res}")

        if mode == self.ModeAnd:  # either la and lb is empty, mode AND
            info("    mode AND")
            res = lo_res  # will be empty
        elif mode in [self.ModeOr, self.ModeXor]:
            info("    mode OR/XOR")
            res = la_res + lo_res + lb_res
        elif mode == self.ModeANotB:
            info("    mode ANotB")
            res = la_res + lo_res
        elif mode == self.ModeBNotA:
            info("    mode BNotA")
            res = lo_res + lb_res
        else:
            res = []

        res = self.normalize(res)
        # res = self.merge_layers_same_z(res)
        # res = self.merge_layers_same_mask(res)
        # res.sort()
        info(f"    boolean_l2l().res = {res}")
        return res

    def split_layers_z(self, a, b):
        """Split two layers if they overlap in z-direction

        Parameters
        ----------
        a : MaterialLayer
        b : MaterialLayer

        Return
        ------
        res : tuple of MaterialLayer
            ab, ao, at, bb, bo, bt. bottom, overlapping and top part of each
            initial layer.

        """
        if not a.is_z_overlapping(b):
            if a.is_lower(b):
                return a, None, None, None, None, b
            else:
                return None, None, a, b, None, None

        overlap_bottom, overlap_top = a.z_overlap(b)
        if a.bottom < overlap_bottom:
            ab_bottom, ab_top = a.bottom, overlap_bottom
        else:
            ab_bottom, ab_top = None, None

        if a.top > overlap_top:
            at_bottom, at_top = overlap_top, a.top
        else:
            at_bottom, at_top = None, None

        if b.bottom < overlap_bottom:
            bb_bottom, bb_top = b.bottom, overlap_bottom
        else:
            bb_bottom, bb_top = None, None

        if b.top > overlap_top:
            bt_bottom, bt_top = overlap_top, b.top
        else:
            bt_bottom, bt_top = None, None

        ab = MaterialLayer(a.mask, ab_bottom, ab_top - ab_bottom) if ab_bottom else None
        at = MaterialLayer(a.mask, at_bottom, at_top - at_bottom) if at_bottom else None
        ao = MaterialLayer(a.mask, overlap_bottom, overlap_top - overlap_bottom)

        bb = MaterialLayer(b.mask, bb_bottom, bb_top - bb_bottom) if bb_bottom else None
        bt = MaterialLayer(b.mask, bt_bottom, bt_top - bt_bottom) if bt_bottom else None
        bo = MaterialLayer(b.mask, overlap_bottom, overlap_top - overlap_bottom)

        return ab, ao, at, bb, bo, bt

    @print_info(False)
    def size_l2l(self, layers, dx, dy=0, dz=0, mode=2, rh=True, mc=True):
        """Change mask size in each layer by dx and dy.

        Size in z-direction remains unchanged.

        Parameters
        ----------
        layers : list of MaterialLayer
        dx : int
            size increase in x-direction in [dbu]
        dy : int (optional)
            size increase in y-direction in [dbu]
        dz : int (optional)
            size increase in z-direction in [dbu]
        mode : int
        rh : boolean (optional)
        mc : boolean (optional)

        Returns
        -------
        res : layers : list of MaterialLayer
        """
        res = []
        for l in layers:
            sized_polys = self.size_p2p(l.mask.data, dx, dy, mode, rh, mc)
            res.append(
                MaterialLayer(
                    LayoutData(sized_polys, l.mask._xs),
                    l.bottom - dz,
                    l.thickness + 2 * dz,
                )
            )

        # Join overlapping layers
        info(f"    res before normalize = {res}")
        res = self.normalize(res)
        info(f"    res after normalize = {res}")

        return res

    @print_info(False)
    def merge_layers_same_z(self, layers):
        """
        Parameters
        ----------
        layers : list of MaterialLayer

        Returns
        -------
        res : list of MaterialLayer
        """
        n_layers = len(layers)
        res_merged = []

        for i, li in enumerate(layers):
            merged = MaterialLayer(li.mask, li.bottom, li.thickness)
            for j in range(i + 1, n_layers):
                lj = layers[j]
                if merged.is_z_same(lj):

                    # perform OR operation on the LayoutData
                    merged = MaterialLayer(
                        merged.mask.or_(lj.mask), merged.bottom, merged.thickness
                    )
                    info(f"    Merged layers b = {merged.bottom}, t = {merged.top}")
            res_merged.append(merged)
        return res_merged

    @print_info(False)
    def merge_layers_same_mask(self, layers):
        """
        Parameters
        ----------
        layers : list of MaterialLayer

        Returns
        -------
        res : list of MaterialLayer
        """
        _check_layer_list_sorted(layers)

        res_merged = []

        while layers:
            la = layers.pop(0)
            ib = 0
            while layers and (ib < len(layers)):
                lb = layers[ib]
                if la.top == lb.bottom:
                    if la.mask.data == lb.mask.data:
                        la = MaterialLayer(la.mask, la.bottom, lb.top - la.bottom)
                        info(
                            f"    Merged layers ({la.bottom},{la.top}) and ({lb.bottom}, {lb.top})"
                        )
                        layers.pop(ib)
                elif la.top < lb.bottom:
                    # all following lb will be higher
                    res_merged.append(la)
                    break  # go to the next la
                ib += 1
            else:
                res_merged.append(la)
        return res_merged


class MaterialLayer:
    def __init__(self, mask, elevation, thickness):
        """
        Parameters
        ----------
        mask : LayoutData
        elevation : int
            z-coordinate of the layer bottom in [dbu]
        thickness : float
            thickness of the layer in [dbu]
        """
        self.mask = mask
        self._bottom = elevation
        self._top = self._bottom
        self.thickness = thickness

    def __lt__(self, other):
        """
        Parameters
        ----------
        other : MaterialLayer
        """
        if self._bottom < other.bottom:
            return True
        elif self._bottom > other.bottom:
            return False
        else:
            return self._top < other.top

    def __str__(self):
        n_edges = "".join(f"{poly.num_points()}, " for poly in self.mask.data)
        return f"<MatLayer (n_polys={self.mask.n_poly}, n_edges=({n_edges[:-2]}), btm = {self._bottom}, top = {self._top})>"

    def __repr__(self):
        n_edges = "".join(f"{poly.num_points()}, " for poly in self.mask.data)
        return f"<MatLayer (n_polys={self.mask.n_poly}, n_edges=({n_edges[:-2]}), btm = {self._bottom}, top = {self._top})>"

    @property
    def bottom(self):
        return self._bottom

    @property
    def top(self):
        return self._top

    @property
    def thickness(self):
        return self._top - self._bottom

    @thickness.setter
    def thickness(self, t):
        """
        Parameters
        ----------
        t : int
            thickness in [dbu]
        """
        if t <= 0:
            raise ValueError(
                f"Material layer thickness must be positive. {t} is given."
            )
        else:
            self._top = self._bottom + t

    @print_info(False)
    def is_z_overlapping(self, other):
        """Check two layers for overlap.

        Parameters
        ----------
        other : MaterialLayer

        Returns
        -------
        bool
        """
        info(
            f"   self.b, self.t, other.b, other.t = {self.bottom} {self.top} {other.bottom} {other.top}"
        )

        return self._top > other.bottom and self._bottom < other.top

    def is_z_same(self, other):
        return self._bottom == other.bottom and self._top == other.top

    def split(self, z_coords):
        """Split layer into several layers.

        Parameters
        ----------
        z_coords : list of int

        Returns
        -------
        res : list of MaterialLayer
        """
        z_coords_all = [self.bottom] + z_coords + [self.top]
        return [
            MaterialLayer(self.mask, b, t - b)
            for b, t in zip(z_coords_all[:-1], z_coords_all[1:])
        ]

    def split_by_layer(self, other):

        z_split = []
        if self.bottom < other.bottom < self.top:
            z_split.append(other.bottom)

        if self.bottom < other.top < self.top:
            z_split.append(other.top)

        return self.split(z_split)

    def z_overlap(self, other):
        """Return overlap points with the other layer.

        self and other must be overlapping.

        Parameters
        ----------
        other : MaterialLayer

        Return
        ------
        res : tuple
            bottom : int or None
            top : int  or None
        """
        bottom = max(self._bottom, other.bottom)
        top = min(self._top, other.top)
        return bottom, top

    def is_lower_s(self, other, levela="bottom", levelb=None):
        """Compares the location of two layers strictly.

        Parameters
        ----------
        other : MaterialLayer
        levela : str
            'bottom|top'
        levelb : str or None
            'bottom|top'. If None, levela will be used

        Returns
        -------
        bool
        """
        if not levelb:
            levelb = levela

        return getattr(self, levela) < getattr(other, levelb)

    def is_lower(self, other, levela="bottom", levelb=None):
        """Compares the location of two layers nonstrictly.

        Parameters
        ----------
        other : MaterialLayer
        levela : str
            'bottom|top'
        levelb : str or None
            'bottom|top'. If None, levela will be used

        Returns
        -------
        bool
        """
        if not levelb:
            levelb = levela

        return getattr(self, levela) <= getattr(other, levelb)

    def is_higher(self, other, levela="bottom", levelb=None):
        """Compares the tops of two layers strictly.

        Parameters
        ----------
        other : MaterialLayer

        Returns
        -------
        bool
        """
        if not levelb:
            levelb = levela

        return getattr(self, levela) > getattr(other, levelb)


def _check_layer_list_sorted(layers):
    for la, lb in zip(layers[:-1], layers[1:]):
        if (la.bottom > lb.bottom) or ((la.bottom == lb.bottom) and (la.top > lb.top)):
            raise ValueError(
                f"layers must be a sorted list of MaterialLayer. Layers {la} and {lb} are not sorted."
            )


@print_info(False)
def layer_to_tech_str(
    layer_no_gds,
    layer,
    name="",
    color=None,
    filter=0.0,
    metal=0,
    shortcut="",
    show=True,
):
    """
    Parameters
    ----------
    layer_no_gds : int
        layer number in the gds file
    layer : MaterialLayer
        layer to be exported
    name : str
        name to be displayed in the legend. If empty, layer number is used.
    color : tuple of float
        r, g, b components of the color in the range [0, 1] each
    filter : float
        layer transparency, from 0 to 1
    metal : float
        Not used at the moment
    shortcut : str
        A digit from 0 to 9. Defines a shortcut from 0 to 9 to toggle the layer
        visibility. Can be prepended with any combination of <Alt>, <Ctrl>,
        and <Shift> as modifiers (eg. '<Shift> 0')
    show : bool
        Whether to show layer during rendering

    Returns
    -------
    s : str
        A layer record for the tech file of GDS3D software.

    """
    name = f"{name} ({layer_no_gds})" if name else f"-- ({layer_no_gds})"
    if (color is None) or (len(color) not in (3, 4)):
        r, g, b = random(), random(), random()
        a = filter
    elif len(color) == 3:
        r, g, b = color
        a = filter
    elif len(color) == 4:
        r, g, b, a = color

    if not (0 <= r <= 1 and 0 <= g <= 1 and 0 <= b <= 1):
        raise ValueError(
            f"Color components must be from 0 to 1. ({r}, {g}, {b}) is given"
        )

    if not (0 <= a <= 1):
        raise ValueError(
            f"Filter / transparency value must be from 0 to 1. {a} is given"
        )

    s = f"LayerStart: {name}\n"
    s += f"Layer: {layer_no_gds}\n"
    s += f"Height: {layer.bottom}\n"
    s += f"Thickness: {layer.thickness}\n"

    s += f"Red: {r}\nGreen: {g}\nBlue: {b}\nFilter: {a}\n"
    s += f"Metal: {metal}\n"
    s += f"Shortkey: {shortcut}\n" if shortcut else ""
    s += f"Show: {int(show)}\n"
    s += "LayerEnd\n\n"
    return s


LP = LayerProcessor
lp = LayerProcessor()



"""klayout_pyxs.compat.py

This module imports functions necessary to ensure compatibility with
both 2.7 and 3.7 Python versions.

Copyright 2017-2019 Dima Pustakhod

"""
import sys

major, middle, minor, _, _ = sys.version_info

if major == 2:
    from six.moces import zip
    from six.moves import range
elif major == 3:
    range = range
    zip = zip
else:
    raise OSError("Unsupported python version")

__all__ = [
    "range",
    "zip",
]



""" pyxs.geometry_2d.py

(C) 2017-2019 Dima Pustakhod and contributors
"""
import math

from klayout_pyxs import (
    EP_,
    DPoint,
    Edges,
    Point,
    Polygon,
    Region,
    SimplePolygon,
    Trans,
)
from klayout_pyxs.compat import range
from klayout_pyxs.layer_parameters import string_to_layer_info
from klayout_pyxs.utils import info, int_floor, make_iterable, print_info


class EdgeProcessor(EP_):
    """
    Problems: empty polygon arrays produce errors with boolean_to_polygon
    because RBA does not recognize the empty array as an array of polygons
    and then there is an ambiguity between the edge-input and polygon input
    variants. Thus this extension which checks for empty input and performs
    some default operation
    """

    def boolean_p2p(self, pa, pb, mode, rh=True, mc=True):
        """Boolean operation for a set of given polygons, creating polygons

        This method computes the result for the given boolean operation on
        two sets of polygons. This method produces polygons on output and
        allows to fine-tune the parameters for that purpose.

        This is a convenience method that bundles filling of the edges,
        processing with a Boolean operator and puts the result into an output
        vector.

        Parameters
        ----------
        pa : list of Polygon
            the input polygons (first operand)
        pb : list of Polygon
            the input polygons (second operand)
        mode : int
            one of self.ModeANotB, self.ModeAnd, self.ModeBNotA,
            self.ModeOr, self.ModeXor
        rh : bool (optional)
            True, if holes should be resolved into the hull
        mc : bool (optional)
            True, if touching corners should be resolved into less connected
            contours

        Returns
        -------
        res : list of Polygon
            The output polygons

        """
        return super().boolean_p2p(pa, pb, mode, rh, mc)

    def safe_boolean_to_polygon(self, pa, pb, mode, rh=True, mc=True):
        """Applies boolean operation to two lists of polygons.

        Use of this method is deprecated. Use boolean_p2p instead

        Works safe in case any of input arrays is empty.

        Parameters
        ----------
        pa : list of Polygon
        pb : list of Polygon
        mode : int
        rh : bool (optional)
            resolve_holes
        mc : bool (optional)
            min_coherence

        Returns
        -------
        list of Polygon or []
        """
        n_pa, n_pb = len(pa), len(pb)  # number of polygons in pa and pb

        if n_pa > 0 and n_pb > 0:  # both pa and pb are not empty
            return self.boolean_to_polygon(pa, pb, mode, rh, mc)
        elif mode == self.ModeAnd:  # either pa and pb is empty, mode AND
            return []  # will be empty
        elif mode == self.ModeOr:
            return pa if n_pa > 0 else pb
        elif mode == self.ModeXor:
            return pa if n_pa > 0 else pb
        elif mode == self.ModeANotB:
            return pa
        elif mode == self.ModeBNotA:
            return pb
        else:
            return []

    def size_to_polygon(self, polygons, dx, dy, mode=2, rh=True, mc=True):
        """Size the given polygons into polygons

        Use of this method is deprecated. Use size_p2p instead
        """
        return super().size_to_polygon(polygons, dx, dy, mode, rh, mc)

    @print_info(False)
    def size_p2p(self, polygons, dx, dy=0, mode=2, rh=True, mc=True):
        """Size the given polygons into polygons

        Parameters
        ----------
        polygons : list of Polygon
            The input polygons
        dx : int
            The sizing value in x direction in dbu
        dy : int (optional)
            The sizing value in y direction in dbu
        mode : int (optional)
            The sizing mode. Allowed values from 1 to 5
        rh : bool (optional)
            True, if holes should be resolved into the hull
        mc : bool (optional)
            True, if touching corners should be resolved into less connected
            contours

        Returns
        -------
        res : list of Polygon
            The output polygons

        """
        info(f"    polys  = {polygons}")
        info(f"    dx = {dx}, dy = {dy}")
        res = super().size_p2p(polygons, dx, dy, mode, rh, mc)
        info(f"    EP.size_p2p().res = {res}")
        return res


EP = EdgeProcessor
ep = EdgeProcessor()


def parse_grow_etch_args(
    method, material_cls, into=(), through=(), on=(), mode="square"
):
    """
    Parameters
    ----------
    method : str
        'etch|grow': calling method, used for debug messages.
    material_cls : type
        into, through, and on lists must contain instances of this type.
    into : None or list (optional)
    on : None or list (optional)
    through : None or list (optional)
    mode : str (optional)
        'square|round|octagon'


    Returns
    -------
    res : tuple
        into : None or list
        through : None or list
        on : None or list
        mode : str
            'square|round|octagon'

    """
    if into:
        into = make_iterable(into)
        for i in into:
            # should be MaterialData @@@
            if not isinstance(i, material_cls):
                raise TypeError(
                    f"'{method}' method: 'into' expects a material parameter or an array of such. {type(i)} is given"
                )

    if on:
        on = make_iterable(on)
        for i in on:
            # should be MaterialData @@@
            if not isinstance(i, material_cls):
                raise TypeError(
                    f"'{method}' method: 'on' expects a material parameter or an array of such"
                )

    if through:
        through = make_iterable(through)
        for i in through:
            # should be MaterialData @@@
            if not isinstance(i, material_cls):
                raise TypeError(
                    f"'{method}' method: 'through' expects a material parameter or an array of such"
                )

    if on and (through or into):
        raise ValueError(
            "'on' option cannot be combined with 'into' or " "'through' option"
        )

    if mode not in ("round", "square", "octagon"):
        raise ValueError(
            f"'{method}' method: 'mode' should be 'round', 'square' or 'octagon'"
        )

    return into, through, on, mode


class LayoutData:
    """Class to manipulate masks, which is a 2d view.

    Layout data is a list of polygons.

    Attributes
    ----------
    self._polygons : list of Polygon
        In case of XSectionGenerator.layer() object, self._polygons
        contains shapes touching the ruler, top view of the mask
    """

    def __init__(self, polygons, xs):
        """LayoutData constructor.

        Parameters
        ----------
        polygons : list of Polygon
            list of shapes contained in this LayoutData
        xs : XSectionGenerator

        """
        self._polygons = polygons
        self._xs = xs
        self._ep = ep

    def upcast(self, polygons):
        return self.__class__(polygons, self._xs)

    def dup(self):
        return self.__class__(self._polygons, self._xs)

    def __str__(self):
        n_poly = self.n_poly

        s = f"LayoutData (n_polygons = {n_poly})"

        if n_poly > 0:
            s += ":"

        for pi in range(min(2, n_poly)):
            s += f"\n    {self._polygons[pi]}"
        return s

    def __repr__(self):
        return f"<LayoutData (n_polygons = {self.n_poly})>"

    @property
    def data(self):
        """
        Return
        ------
        data: list of Polygon
            polygons which constitute the mask
        """
        return self._polygons

    @data.setter
    def data(self, polygons):
        """
        Parameters
        ----------
        polygons: list of Polygon
            polygons to be saved in the mask
        """
        self._polygons = polygons

    def add(self, other):
        """Add more polygons to the layout (OR).

        Parameters
        ----------
        other : LayoutData or list of Polygon

        """
        other_polygons = self._get_polygons(other)
        self._polygons = self._ep.boolean_p2p(self._polygons, other_polygons, EP.ModeOr)

    def and_(self, other):
        """Calculate overlap of the mask with a list of polygons (AND).

        Parameters
        ----------
        other : LayoutData or list of Polygon

        Returns
        -------
        ld : LayoutData
        """
        other_polygons = self._get_polygons(other)
        return self.upcast(
            self._ep.boolean_p2p(self._polygons, other_polygons, EP.ModeAnd)
        )

    def invert(self):
        self._polygons = self._ep.boolean_p2p(
            self._polygons, [Polygon(self._xs.background())], EP.ModeXor
        )

    def inverted(self):
        """Calculate inversion of the mask.

        Total region is determined by self._xs.background().

        Returns
        -------
        ld : LayoutData
        """
        return self.upcast(
            self._ep.boolean_p2p(
                self._polygons, [Polygon(self._xs.background())], EP.ModeXor
            )
        )

    @print_info(False)
    def load(self, layout, cell, box, layer_spec):
        """Load all shapes from the layer into self._polygons.

        The shapes are collected from layer defined by layer_spec. Only
        shapes touching the box are loaded. Box is effectively a ruler region.

        Parameters
        ----------
        layout : Layout
            layout
        cell : int
            cell's index
        box : Box
            The box of the ruler, enlarged in both directions.
            Only shapes touching this box will be collected
        layer_spec : str
            layer to be used
        """
        info(f"LD.load(..., box={box}, layer_spec={layer_spec})")

        ls = string_to_layer_info(layer_spec)

        # look up the layer index with a given layer_spec in the current layout
        layer_index = None
        for li in layout.layer_indices():
            info(f"    li = {li}")
            if layout.get_info(li).is_equivalent(ls):
                info(f"        layer_index = {li}")
                layer_index = li
                break

        # collect polygons from the specified layer
        # all the shapes from the layout will be saved in self._polygons
        if layer_index is not None:
            info("    iterations:")
            shape_iter = layout.begin_shapes_touching(cell, layer_index, box)

            while not shape_iter.at_end():
                shape = shape_iter.shape()
                if shape.is_polygon() or shape.is_path() or shape.is_box():
                    self._polygons.append(
                        shape.polygon.transformed(shape_iter.itrans())
                    )
                shape_iter.next()

        n_poly = self.n_poly
        info(f"    loaded polygon count: {n_poly}")
        if n_poly > 0:
            info("    loaded polygons:")
        for pi in range(min(2, n_poly)):
            info(f"        {self._polygons[pi]}")

        info("LD.load()\n")

    def mask(self, other):
        """Mask current layout with external list of polygons (AND).

        Parameters
        ----------
        other : LayoutData or list of Polygon

        """
        other_polygons = self._get_polygons(other)
        self._polygons = self._ep.boolean_p2p(
            self._polygons, other_polygons, EP.ModeAnd
        )

    @property
    def n_poly(self):
        """
        Returns
        -------
        n_poly : int
            number of polygons contained in the mask

        """
        return len(self._polygons)

    def not_(self, other):
        """Calculate difference with another list of polygons.

        Parameters
        ----------
        other : LayoutData or list of Polygon

        Returns
        -------
        ld : LayoutData
        """
        other_polygons = self._get_polygons(other)
        return self.upcast(
            self._ep.boolean_p2p(self._polygons, other_polygons, EP.ModeANotB)
        )

    __sub__ = not_

    def or_(self, other):
        """Calculate sum with another list of polygons (OR).

        Parameters
        ----------
        other : LayoutData or list of Polygon

        Returns
        -------
        ld : LayoutData
        """
        other_polygons = self._get_polygons(other)
        return self.upcast(
            self._ep.boolean_p2p(self._polygons, other_polygons, EP.ModeOr)
        )

    __add__ = or_
    __iadd__ = or_

    def size(self, dx, dy=None):
        """Resize the layout mask.

        Parameters
        ----------
        dx : float
            size change in x-direction in [um]
        dy : float (optional)
            size change in y-direction in [um]. Equals to dx by default.

        """
        dy = dx if dy is None else dy
        self._polygons = self._ep.size_p2p(
            self._polygons,
            int_floor(dx / self._xs.dbu + 0.5),
            int_floor(dy / self._xs.dbu + 0.5),
        )

    def sized(self, dx, dy=None):
        """Calculate a sized mask.

        Parameters
        ----------
        dx : float
            size change in x-direction in [um]
        dy : float (optional)
            size change in y-direction in [um]. Equals to dx by default.

        Returns
        -------
        ld : LayoutData
        """
        dy = dx if dy is None else dy
        return self.upcast(
            self._ep.size_p2p(
                self._polygons,
                int_floor(dx / self._xs.dbu + 0.5),
                int_floor(dy / self._xs.dbu + 0.5),
            )
        )

    def sub(self, other):
        """Subtract another list of polygons.

        Parameters
        ----------
        other : LayoutData or list of Polygon

        """
        other_polygons = self._get_polygons(other)
        self._polygons = self._ep.boolean_p2p(
            self._polygons, other_polygons, EP.ModeANotB
        )

    def transform(self, t):
        """Transform mask with a transformation.

        Parameters
        ----------
        t : Trans
            transformation to be applied
        """
        self._polygons = [p.transformed(t) for p in self._polygons]

    def xor(self, other):
        """Calculate XOR with another list of polygons.

        Parameters
        ----------
        other : LayoutData or list of Polygon

        Returns
        -------
        ld : LayoutData
        """
        other_polygons = self._get_polygons(other)
        return self.upcast(
            self._ep.boolean_p2p(self._polygons, other_polygons, EP.ModeXor)
        )

    def close_gaps(self):
        """Close gaps in self._polygons.

        Increase size of all polygons by 1 dbu in all directions.
        """
        sz = 1
        d = self._polygons
        d = self._ep.size_p2p(d, 0, sz)
        d = self._ep.size_p2p(d, 0, -sz)
        d = self._ep.size_p2p(d, sz, 0)
        d = self._ep.size_p2p(d, -sz, 0)
        self._polygons = d

    def remove_slivers(self):
        """Remove slivers in self._polygons."""
        sz = 1
        d = self._polygons
        d = self._ep.size_p2p(d, 0, -sz)
        d = self._ep.size_p2p(d, 0, sz)
        d = self._ep.size_p2p(d, -sz, 0)
        d = self._ep.size_p2p(d, sz, 0)
        self._polygons = d

    @staticmethod
    def _get_polygons(l):
        if isinstance(l, LayoutData):
            return l.data
        elif isinstance(l, (tuple, list)):
            return l
        else:
            raise TypeError(
                f"l should be either an instance of LayoutData or a list of Polygon. {type(l)} is given."
            )


class MaskData(LayoutData):
    """Class to operate 2D cross-sections.

    Material data is a list of single

    """

    @print_info(False)
    def __init__(self, air_polygons, mask_polygons, xs):
        """
        Parameters
        ----------
        air_polygons : list of Polygon
            list of shapes constituting air in cross-section
        mask_polygons : list of Polygon
            list of shapes constituting material in cross-section
        xs: XSectionGenerator
            passed to LayoutData.__init__()
        delta : float
            the intrinsic height (required for mask data because there
            cannot be an infinitely small mask layer (in database units)
        """
        super().__init__([], xs)  # LayoutData()
        self._air_polygons = air_polygons
        self._mask_polygons = mask_polygons

        info(f"air_polygons = {air_polygons}")
        info(f"mask_polygons = {mask_polygons}")
        info("Success!")

    def upcast(self, polygons):
        return MaskData(self._air_polygons, polygons, self._xs)

    def dup(self):
        return MaskData(self._air_polygons, self._mask_polygons, self._xs)

    def __str__(self):
        n_air_poly = self.n_air_poly
        n_mask_poly = self.n_mask_poly

        s = f"{self.__class__.__name__} (n_air_polygons={n_air_poly}, n_mask_polygons={n_mask_poly})"

        if n_mask_poly > 0:
            s += ":"

        for pi in range(min(2, n_mask_poly)):
            s += f"\n    {self._mask_polygons[pi]}"
        return s

    @property
    def n_air_poly(self):
        """
        Returns
        -------
        int
            number of polygons describing air

        """
        return len(self._air_polygons)

    @property
    def n_mask_poly(self):
        """
        Returns
        -------
        int
            number of polygons describing mask

        """
        return len(self._mask_polygons)

    def __repr__(self):
        return f"<MaskData (delta={self._delta}, n_air_polygons={self.n_air_poly}, n_mask_polygons={self.n_mask_poly})>"

    @print_info(False)
    def grow(
        self,
        z,
        xy=0.0,
        into=(),
        through=(),
        on=(),
        mode="square",
        taper=None,
        bias=None,
        buried=None,
    ):
        """
        Parameters
        ----------
        z : float
            height
        xy : float
            lateral
        mode : str
            'round|square|octagon'. The profile mode.
        taper : float
            The taper angle. This option specifies tapered mode and cannot
            be combined with :mode.
        bias : float
            Adjusts the profile by shifting it to the interior of the figure.
            Positive values will reduce the line width by twice the value.
        on : list of MaterialData (optional)
            A material or an array of materials onto which the material is
            deposited (selective grow). The default is "all". This option
            cannot be combined with ":into". With ":into", ":through" has the
            same effect than ":on".
        into : list of MaterialData (optional)
            Specifies a material or an array of materials that the new
            material should consume instead of growing upwards. This will
            make "grow" a "conversion" process like an implant step.
        through : list of MaterialData (optional)
            To be used together with ":into". Specifies a material or an array
            of materials to be used for selecting grow. Grow will happen
            starting on the interface of that material with air, pass
            through the "through" material (hence the name) and consume and
            convert the ":into" material below.
        buried : float
            Applies the conversion of material at the given depth below the
            mask level. This is intended to be used together with :into
            and allows modeling of deep implants. The value is the depth
            below the surface.

        """
        # parse the arguments
        info(f"    into={into}")
        into, through, on, mode = parse_grow_etch_args(
            "grow", MaterialData, into=into, through=through, on=on, mode=mode
        )

        info(f"    into={into}")
        # produce the geometry of the new material
        d = self.produce_geom(
            "grow", xy, z, into, through, on, taper, bias, mode, buried
        )

        # prepare the result
        # list of Polygon which are removed
        res = MaterialData(d, self._xs)

        # consume material
        if into:
            for i in into:  # for each MaterialData
                i.sub(res)
        else:
            self._xs.air().sub(res)  # remove air where material was added
        return res

    def etch(
        self,
        z,
        xy=0.0,
        into=(),
        through=(),
        mode="square",
        taper=None,
        bias=None,
        buried=None,
    ):
        """

        Parameters
        ----------
        z : float
            etch depth
        xy : float (optional)
            mask extension, lateral
        mode : str
            'round|square|octagon'. The profile mode.
        taper :	float
            The taper angle. This option specifies tapered mode and cannot
            be combined with mode.
        bias : float
            Adjusts the profile by shifting it to the interior of the
            figure. Positive values will reduce the line width by twice
            the value.
        into :	list of MaterialData (optional)
            A material or an array of materials into which the etch is
            performed. This specification is mandatory.
        through : list of MaterialData (optional)
            A material or an array of materials which form the selective
            material of the etch. The etch will happen only where this
            material interfaces with air and pass through this material
            (hence the name).
        buried : float
            Applies the etching at the given depth below the surface. This
            option allows to create cavities. It specifies the vertical
            displacement of the etch seed and there may be more applications
            for this feature.

        """
        # parse the arguments

        into, through, on, mode = parse_grow_etch_args(
            "etch", MaterialData, into=into, through=through, on=(), mode=mode
        )

        if not into:
            raise ValueError("'etch' method: requires an 'into' specification")

        # prepare the result
        d = self.produce_geom(
            "etch", xy, z, into, through, on, taper, bias, mode, buried
        )  # list of Polygon

        # produce the geometry of the etched material
        # list of Polygon which are removed
        res = MaterialData(d, self._xs)

        # consume material and add to air
        for i in into:  # for each MaterialData
            j = LayoutData(i.data, self._xs)
            i.sub(res)
            j.sub(i)
            self._xs.air().add(j)

        # Add air in place of the etched materials
        # self._xs.air().add(res)
        # self._xs.air().close_gaps()

    @print_info(False)
    def produce_geom(self, method, xy, z, into, through, on, taper, bias, mode, buried):
        """

        Parameters
        ----------
        method : str
        xy : float
            extension
        z : float
            height
        into : list of MaterialData
        through : list of MaterialData
        on : list of MaterialData
        taper : float
        bias : float
        mode : str
            'round|square|octagon'
        buried :

        Returns
        -------
        d : list of Polygon
        """
        info(f"    method={method}, xy={xy}, z={z},")
        info(f"    into={into}, through={through}, on={on},")
        info(
            "    taper={}, bias={}, mode={}, buried={})".format(
                taper, bias, mode, buried
            )
        )

        prebias = bias or 0.0

        if xy < 0.0:  # if size to be reduced,
            xy = -xy  #
            prebias += xy  # positive prebias

        if taper:
            d = z * math.tan(math.pi / 180.0 * taper)
            prebias += d - xy
            xy = d

        # determine the "into" material by joining the data of all "into" specs
        # or taking "air" if required.
        # into_data is a list of polygons from all `into` MaterialData
        # Finally we get a into_data, which is a list of Polygons
        if into:
            into_data = []
            for i in into:
                if len(into_data) == 0:
                    into_data = i.data
                else:
                    into_data = self._ep.boolean_p2p(i.data, into_data, EP.ModeOr)
        else:
            # when deposit or grow is selected, into_data is self.air()
            into_data = self._xs.air().data

        info(f"    into_data = {into_data}")

        # determine the "through" material by joining the data of all
        # "through" specs
        # through_data is a list of polygons from all `through` MaterialData
        # Finally we get a through_data, which is a list of Polygons
        if through:
            through_data = []
            for i in through:
                if len(through_data) == 0:
                    through_data = i.data
                else:
                    through_data = self._ep.boolean_p2p(i.data, through_data, EP.ModeOr)
            info(f"    through_data = {through_data}")

        # determine the "on" material by joining the data of all "on" specs
        # on_data is a list of polygons from all `on` MaterialData
        # Finally we get an on_data, which is a list of Polygons
        if on:
            on_data = []
            for i in on:
                if len(on_data) == 0:
                    on_data = i.data
                else:
                    on_data = self._ep.boolean_p2p(i.data, on_data, EP.ModeOr)
            info(f"    on_data = {on_data}")

        pi = int_floor(prebias / self._xs.dbu + 0.5)
        xyi = int_floor(xy / self._xs.dbu + 0.5)
        zi = int_floor(z / self._xs.dbu + 0.5)

        # calculate all edges without prebias and check if prebias
        # would remove edges if so reduce it
        mp = self._ep.size_p2p(self._mask_polygons, 0, 0, 2)

        for p in mp:
            box = p.bbox()
            if box.width() <= 2 * pi:
                pi = int_floor(box.width() / 2.0) - 1
                xyi = pi

        mp = self._ep.size_p2p(self._mask_polygons, -pi, 0, 2)
        air_masked = self._ep.boolean_p2p(self._air_polygons, mp, EP.ModeAnd)
        me = (Edges(air_masked) if air_masked else Edges()) - (
            Edges(mp) if mp else Edges()
        )
        info(f"me after creation: {me}")

        # in the "into" case determine the interface region between
        # self and into
        if into or through or on:
            if on:
                data = on_data
            elif through:
                data = through_data
            else:
                data = into_data

            info(f"data = {data}")
            me = (me & Edges(data)) if data else list()

            # if len(data) == 0:
            #     me = []
            # else:
            #     me += Edges(data)
        info(f"type(me): {type(me)}")  # list of Edge
        info(f"me before operation: {me}")

        d = Region()

        if taper and xyi > 0:
            info("    case taper and xyi > 0")
            kernel_pts = list()
            kernel_pts.append(Point(-xyi, 0))
            kernel_pts.append(Point(0, zi))
            kernel_pts.append(Point(xyi, 0))
            kernel_pts.append(Point(0, -zi))
            kp = Polygon(kernel_pts)
            for e in me:
                d.insert(kp.minkowsky_sum(e, False))

        elif xyi <= 0:
            info("    case xyi <= 0")
            # TODO: there is no way to do that with a Minkowsky sum currently
            # since polygons cannot be lines except through dirty tricks
            dz = Point(0, zi)
            for e in me:
                d.insert(Polygon([e.p1 - dz, e.p2 - dz, e.p2 + dz, e.p1 + dz]))
        elif mode in ("round", "octagon"):
            info("    case round / octagon")
            # approximate round corners by 64 points for "round" and
            # 8 for "octagon"
            n = 64 if mode == "round" else 8
            da = 2.0 * math.pi / n
            rf = 1.0 / math.cos(da * 0.5)

            info(f"    n = {n}, da = {da}, rf = {rf}")
            kernel_pts = list()
            for i in range(n):
                kernel_pts.append(
                    Point.from_dpoint(
                        DPoint(
                            xyi * rf * math.cos(da * (i + 0.5)),
                            zi * rf * math.sin(da * (i + 0.5)),
                        )
                    )
                )
            info(f"    n kernel_pts: {len(kernel_pts)}")
            info(f"    kernel_pts: {kernel_pts}")

            kp = Polygon(kernel_pts)
            for n, e in enumerate(me):
                d.insert(kp.minkowsky_sum(e, False))
                if n > 0 and n % 10 == 0:
                    d.merge()

        elif mode == "square":
            kernel_pts = list()
            kernel_pts.append(Point(-xyi, -zi))
            kernel_pts.append(Point(-xyi, zi))
            kernel_pts.append(Point(xyi, zi))
            kernel_pts.append(Point(xyi, -zi))
            kp = SimplePolygon()
            kp.set_points(kernel_pts, True)  # "raw" - don't optimize away
            for e in me:
                d.insert(kp.minkowsky_sum(e, False))

        d.merge()
        info(f"d after merge: {d}")

        if abs(buried or 0.0) > 1e-6:
            t = Trans(Point(0, -int_floor(buried / self._xs.dbu + 0.5)))
            d.transform(t)
        if through:
            d -= Region(through_data)
        d &= Region(into_data)

        poly = [p for p in d]
        return poly


class MaterialData(LayoutData):
    def __init__(self, polygons, xs):
        super().__init__(polygons, xs)

    def discard(self):
        self._xs.air().add(self)

    def keep(self):
        self._xs.air().sub(self)

    def __repr__(self):
        n_poly = self.n_poly

        s = f"{self.__class__.__name__} (n_polygons = {n_poly})"

        if n_poly > 0:
            s += ":"

        for pi in range(min(2, n_poly)):
            s += f"\n    {self._polygons[pi]}"
        return s

    def __str__(self):
        s = f"<{self.__class__.__name__} (n_polygons = {self.n_poly})>"
        return s



"""

Copyright 2017-2019 Dima Pustakhod


Changelog
---------
2019.10.01
    Fix imports from pya/klayout
    Add doctests
2017.xx.xx
    Initial commit
"""
import re

from klayout_pyxs import LayerInfo


def string_to_layer_info_params(layer_spec, return_None=False):
    """Convert the layer specification into a LayerInfo parameters

    Parameters
    ----------
    layer_spec : str
        format: "l", "l/d", "n(l/d)" or "n".

    Returns
    -------
    res : tuple
        layer (int), data type (int), name (str)

    Examples
    --------
    >>> print(string_to_layer_info_params('1'))
    (1, 0)
    >>> print(string_to_layer_info_params('1/2'))
    (1, 2)
    >>> print(string_to_layer_info_params('a(1/2)'))
    (1, 2, 'a')
    >>> print(string_to_layer_info_params('a'))
    ('a',)
    """
    if re.match(r"^(\d+)$", layer_spec):
        match = re.match(r"^(\d+)$", layer_spec)
        ls = int(match[0]), 0
    elif re.match(r"^(\d+)/(\d+)$", layer_spec):
        match = re.match(r"^(\d+)/(\d+)$", layer_spec)
        ls = int(match[1]), int(match[2])
    elif re.match(r"^(.*)\s*\((\d+)/(\d+)\)$", layer_spec):
        match = re.match(r"^(.*)\s*\((\d+)/(\d+)\)$", layer_spec)
        ls = int(match[2]), int(match[3]), match[1]
    else:
        ls = (layer_spec,)

    if return_None:
        if len(ls) == 1:
            ls = (None, None, ls[0])
        elif len(ls) == 2:
            ls = (ls[0], ls[1], None)

    return ls


def string_to_layer_info(layer_spec):
    """Convert the layer specification into a LayerInfo structure

    Parameters
    ----------
    layer_spec : str
        format: "l", "l/d", "n(l/d)" or "n".

    Returns
    -------
    ls : LayerInfo
        layer parameters are given by the `layer_spec`.

    Examples
    --------
    >>> string_to_layer_info('1')
    1/0
    >>> string_to_layer_info('1/2')
    1/2
    >>> string_to_layer_info('a(1/2)')
    a (1/2)
    >>> string_to_layer_info('a')
    a
    """
    ls_param = string_to_layer_info_params(layer_spec)
    return LayerInfo(*ls_param)


def main():
    import doctest

    doctest.testmod()


if __name__ == "__main__":
    main()



from setuptools import find_packages
from setuptools import setup


with open("README.md") as f:
    LONG_DESCRIPTION = f.read()


# def get_install_requires():
#     with open("requirements.txt") as f:
#         return [line.strip() for line in f.readlines() if not line.startswith("-")]


setup(
    name="klayout_pyxs",
    version="0.1.8",
    url="https://github.com/dimapu/klayout_pyxs",
    license="MIT",
    author="Dima Pustakhod",
    description="python port of the Klayout xsection project",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=("tests",)),
    # install_requires=get_install_requires(),
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python",
    ],
)



import math

VERBOSE = True
OFFSET = 0


def info(*s):
    """Print information with offset.

    Parameters
    ----------
    s : str
        string to be printed
    """
    if VERBOSE:
        print(" " * OFFSET, *s)


def print_info(v=True):
    """Decorator to show / disable function output to the console.

    Parameters
    ----------
    v : bool
        it False, all info() inside the function will be disabled.

    """

    def decorator(f):
        def wrapper(*args, **kwargs):
            global VERBOSE
            global OFFSET
            old_v = VERBOSE
            VERBOSE = v
            if v:
                OFFSET += 4
            info(f"{f.__name__}():")
            res = f(*args, **kwargs)
            info(f"end of {f.__name__}()\n")
            if v:
                OFFSET -= 4
            VERBOSE = old_v
            return res

        return wrapper

    return decorator


def int_floor(x):
    """Floor a float value and return int

    Returns
    -------
    res : int
        int(math.floor(x))

    Examples
    --------
    >>> int_floor(1.5)
    1
    >>> int_floor(1.2)
    1
    >>> int_floor(1.8)
    1
    >>> int_floor(-1.2)
    -2
    >>> int_floor(-1.5)
    -2
    >>> int_floor(-1.8)
    -2
    """
    return int(math.floor(x))


def _check_type(instance, typ, caller=""):
    """Check type of an object

    Parameters
    ----------
    caller : str
        caller name. Used for more informative error messages.
    """

    if not isinstance(instance, typ):
        caller_str = f"'{caller}': " if caller != "" else ""
        raise TypeError(
            f"{caller_str}Argument must be an instance of {typ}. {type(instance)} is given"
        )


def make_iterable(v):
    return v if (v is None) or (type(v) in (list, tuple)) else [v]


def main():
    import doctest

    doctest.testmod()


if __name__ == "__main__":
    main()
